import hashlib
import base64
import json
from enum import Enum
from pathlib import Path
from tritondse.types import PathLike, SymExType
from typing import List, Dict, Union, Optional
from dataclasses import dataclass, field
import enum_tools.documentation


@enum_tools.documentation.document_enum
class SeedStatus(Enum):
    """
    Seed status enum.
    Enables giving a status to a seed during its execution.
    At the end of a :py:obj:`SymbolicExecutor` run one of these
    status must have set to the seed.
    """
    NEW = 0      # doc: The input seed is new (has not been executed yet)
    OK_DONE = 1  # doc: The input seed has been executed and terminated correctly
    CRASH = 2    # doc: The input seed crashed in some ways
    HANG = 3     # doc: The input seed made the program to hang
    FAIL = 4     # doc: The input seed made the symbolic executor to raise an exception


@enum_tools.documentation.document_enum
class SeedFormat(Enum):
    """
    Seed format enum
    Raw seeds are just bytes Seed(b"AAAAA\x00BBBBB")
    Composite can describe how to inject the input more precisely 
    """
    RAW = 0        # doc: plain bytes input seed
    COMPOSITE = 1  # doc: complex input object


@dataclass(frozen=True)
class CompositeData:
    argv: List[bytes] = field(default_factory=list)
    "list of argv values"
    files: Dict[str, bytes] = field(default_factory=dict)
    "dictionnary of files and the associated content (stdin is one of them)"
    variables: Dict[str, bytes] = field(default_factory=dict)
    "user defined variables, that the use must take care to inject at right location"

    def _to_json(self):
        data = {
            'argv': [base64.b64encode(v).decode() for v in self.argv],
            'files': {k: (base64.b64encode(v).decode() if isinstance(v, bytes) else v) for k, v in self.files.items()},
            'variables': {k: (base64.b64encode(v).decode() if isinstance(v, bytes) else v) for k, v in self.variables.items()},
        }
        return json.dumps(data, indent=2)

    def __bytes__(self) -> str:
        """
        Serialize data into a json string.

        :return: JSON serialized data
        """
        return self._to_json().encode()

    @staticmethod
    def from_dict(json_data: dict) -> 'CompositeData':
        """
        Convert dict data into a :py:obj:`CompositeData` object.

        :param json_data: json data
        :return: new object instance
        """
        argv = [base64.b64decode(v) for v in json_data['argv']]
        files = {k: (base64.b64decode(v) if isinstance(v, str) else v) for k, v in json_data['files'].items()}
        variables = {k: (base64.b64decode(v) if isinstance(v, str) else v) for k, v in json_data['variables'].items()}
        return CompositeData(argv=argv, files=files, variables=variables)

    def __hash__(self):
        return hash(bytes(self))


class Seed(object):
    """
    Seed input.
    Holds the bytes buffer of the content a status after execution
    but also some metadata of code portions it is meant to cover.
    """
    def __init__(self, content: Union[bytes, CompositeData] = bytes(), status=SeedStatus.NEW):
        """
        :param content: content of the input. By default is b"" *(and is thus considered as a bootstrap seed)*
        :type content: bytes
        :param status: status of the seed if already known
        :type status: SeedStatus
        """
        self.content = content
        self.coverage_objectives = set()  # set of coverage items that the seed is meant to cover
        self.meta_fname = []
        self.target = None                # CovItem informational field indicate the item the seed was generated for
        self._status = status
        self._type = SeedFormat.COMPOSITE if isinstance(content, CompositeData) else SeedFormat.RAW

    def is_composite(self) -> bool:
        """Returns wether the seed is a composite seed or not. """
        return self._type == SeedFormat.COMPOSITE

    def is_raw(self) -> bool:
        """Returns wether the seed is a raw seed or not. """
        return self._type == SeedFormat.RAW

    def is_bootstrap_seed(self) -> bool:
        """
        A bootstrap seed is an empty seed (b""). It will received a
        specific processing in the engine as its size will be automatically
        adapted to the size read (in stdin for instance)

        :return: true if the seed is a bootstrap seed
        """
        return self.content == b""

    def is_fresh(self) -> bool:
        """
        A fresh seed is never been executed. Its is recognizable
        as it does not contain any coverage objectives.

        :return: True if the seed has never been executed
        """
        return not self.coverage_objectives

    @property
    def status(self) -> SeedStatus:
        """
        Status of the seed.

        :rtype: SeedStatus
        """
        return self._status

    @property
    def format(self) -> SeedFormat:
        """
        Format of the seed.

        :rtype: SeedFormat
        """
        return self._type

    @status.setter
    def status(self, value: SeedStatus) -> None:
        """ Sets the status of the seed """
        self._status = value

    def is_status_set(self) -> bool:
        """ Checks whether a status has already been assigned to the seed. """
        return self.status != SeedStatus.NEW

    def __len__(self) -> int:
        """
        Size of the content of the seed.

        :rtype: int
        """
        return len(bytes(self.content))

    def __eq__(self, other) -> bool:
        """
        Equality check based on content.

        :returns: true if content of both seeds are equal """
        return self.content == other.content

    def bytes(self) -> bytes:
        return bytes(self)

    def __bytes__(self) -> bytes:
        """
        Return a representation of the seed's content in bytes.

        :rtype: bytes
        """
        return bytes(self.content)

    def __hash__(self):
        """
        Seed hash function overriden to base itself on content.
        That enable storing seed in dictionnaries directly based
        on their content to discriminate them.

        :rtype: int
        """
        return hash(self.content)

    @property
    def hash(self) -> str:
        """
        MD5 hash of the seed content

        :rtype: str
        """
        m = hashlib.md5(bytes(self))
        return m.hexdigest()

    @property
    def size(self) -> int:
        """
        Size of the seed content in bytes

        :rtype: int
        """
        return len(bytes(self))

    @property
    def filename(self):
        """
        Standardized filename based on hash and size.
        That does not mean the file exists or anything.

        :returns: formatted intended filename of the seed
        :rtype: str
        """
        return f"{self.hash}_{self.size:04x}_{'_'.join(self.meta_fname)}.tritondse.cov"

    @staticmethod
    def from_bytes(raw_seed: bytes, status: SeedStatus = SeedStatus.NEW) -> 'Seed':
        """
        Parse a seed from its byte representation. If its a composite one
        it will parse the bytes as JSON and create the CompositeData accordingly.

        :param raw_seed: bytes: raw bytes of the seed
        :param status: status of the seed if any, otherwise :py:obj:`SeedStatus.NEW`
        :type status: SeedStatus

        :returns: fresh seed instance
        :rtype: Seed
        """
        try:
            data = json.loads(raw_seed)

            if not isinstance(data, dict):  # it might happen that files contains only digit which is a valid JSON
                return Seed(raw_seed, status)

            if 'files' in data and 'argv' in data:
                return Seed(CompositeData.from_dict(data), status)
            else:  # Else still consider file as raw bytes
                return Seed(raw_seed, status)
        except (json.JSONDecodeError, UnicodeDecodeError):
            return Seed(raw_seed, status)

    @staticmethod
    def from_file(path: PathLike, status: SeedStatus = SeedStatus.NEW) -> 'Seed':
        """
        Read a seed from a file. The status can optionally given
        as it cannot be determined from the file.

        :param path: seed path
        :type path: :py:obj:`tritondse.types.PathLike`
        :param status: status of the seed if any, otherwise :py:obj:`SeedStatus.NEW`
        :type status: SeedStatus

        :returns: fresh seed instance
        :rtype: Seed
        """
        raw = Path(path).read_bytes()

        seed = Seed.from_bytes(raw, status)

        # Parse filename to extract back metadata if any
        name = Path(path).name
        if name.endswith(".tritondse.cov"):
            name = name.replace(".tritondse.cov", "")
        metas = name.split("_")
        if len(metas) >= 4:
            seed.meta_fname = metas[3:]

        return seed

    # Utility function for composite seeds
    def is_file_defined(self, name: str) -> bool:
        if self.is_composite():
            return name in self.content.files
        else:
            return False

    def get_file_input(self, name: str) -> bytes:
        """
        Return the bytes associated to a given file within
        a composite seed.

        :raise KeyError: if the name cannot be found in the seed.
        :param name: name of the file to retrieve
        :return: bytes of the file content
        """
        return self.content.files[name]



import logging
import json
from enum import Enum, IntFlag
from pathlib import Path
from typing import List
from functools import reduce

# triton-based libraries
from tritondse.coverage import CoverageStrategy, BranchSolvingStrategy
from tritondse.types import SmtSolver
from tritondse.seed import SeedFormat


class Config(object):
    """
    Data class holding tritondse configuration
    parameter
    """

    def __init__(self,
                 seed_format: SeedFormat = SeedFormat.RAW,
                 pipe_stdout: bool = False,
                 pipe_stderr: bool = False,
                 skip_sleep_routine: bool = False,
                 smt_solver: SmtSolver = SmtSolver.Z3,
                 smt_timeout: int = 5000,
                 execution_timeout: int = 0,
                 exploration_timeout: int = 0,
                 exploration_limit: int = 0,
                 thread_scheduling: int = 200,
                 smt_queries_limit: int = 1200,
                 smt_enumeration_limit: int = 40,
                 coverage_strategy: CoverageStrategy = CoverageStrategy.BLOCK,
                 branch_solving_strategy: BranchSolvingStrategy = BranchSolvingStrategy.FIRST_LAST_NOT_COVERED,
                 workspace: str = "",
                 program_argv: List[str] = None,
                 time_inc_coefficient: float = 0.00001,
                 skip_unsupported_import: bool = False,
                 skip_unsupported_instruction: bool = False,
                 memory_segmentation: bool = True):

        self.seed_format: SeedFormat = seed_format
        """ Seed type is either Raw (raw bytes) or Composite (more expressive).
            See seeds.py for more information on each format.
        """
        
        self.pipe_stdout: bool = pipe_stdout
        """ Pipe the program stdout to Python's stdout. *(default: False)*"""

        self.pipe_stderr: bool = pipe_stderr
        """ Pipe the program stderr to Python's stderr *(default: False)*"""

        self.skip_sleep_routine: bool = skip_sleep_routine
        """ Whether to emulate sleeps routine or not *(default: False)*"""

        self.smt_solver: SmtSolver = smt_solver
        """ SMT solver to perform queries solving """

        self.smt_timeout: int = smt_timeout
        """ Timeout for a single SMT query in milliseconds *(default: 10)*"""

        self.execution_timeout: int = execution_timeout
        """ Timeout of a single execution. If it is triggered the associated
        input file is marked as 'hanging'. In seconds, 0 means unlimited *(default: 0)*"""

        self.exploration_timeout: int = exploration_timeout
        """ Overall timeout of the exploration in seconds. 0 means unlimited *(default: 0)* """

        self.exploration_limit: int = exploration_limit
        """ Number of execution iterations. 0 means unlimited *(default: 0)*"""

        self.thread_scheduling: int = thread_scheduling
        """ Number of instructions to execute before switching to the next thread.
        At the moment all threads are scheduled in a round-robin manner *(default: 200)*"""

        self.smt_queries_limit: int = smt_queries_limit
        """ Limit of SMT queries to perform for a single execution *(default: 1200)*"""

        self.smt_enumeration_limit: int = smt_enumeration_limit
        """ Limit of model values retrieved when enumerating a dynamic jump or symbolic memory accesses"""

        self.coverage_strategy: CoverageStrategy = coverage_strategy
        """ Coverage strategy to apply for the whole exploration, default: :py:obj:`CoverageStrategy.BLOCK`"""

        self.branch_solving_strategy: BranchSolvingStrategy = branch_solving_strategy
        """ Branch solving strategy to apply for a single execution. For a given non-covered
        branch allows changing whether we try to solve it at all occurences or more seldomly.
        default: :py:obj:`BranchSolvingStrategy.FIRST_LAST_NOT_COVERED`
        """

        self.workspace: str = workspace
        """ Workspace directory to use. *(default: 'workspace')* """

        self.program_argv: List[str] = [] if program_argv is None else program_argv
        """ Concrete program argument as given on the command line."""

        self.time_inc_coefficient: float = time_inc_coefficient
        """ Time increment coefficient at each instruction to provide a deterministic
        behavior when calling time functions (e.g gettimeofday(), clock_gettime(), ...).
        For example, if 0.0001 is defined, each instruction will increment the time representation
        of the execution by 100us. *(default: 0.00001)*
        """

        self.skip_unsupported_import: bool = skip_unsupported_import
        """ Whether or not to stop the emulation when hitting a external
        call to a function that is not supported.
        """

        self.skip_unsupported_instruction: bool = skip_unsupported_instruction
        """ Whether or not to stop the emulation when hitting an instruction
        for which the semantic is not defined.
        """

        self.memory_segmentation: bool = memory_segmentation
        """ This option defines whether or not memory segmentation is enforced.
        If activated all memory accesses must belong to a mapped memory area.
        """

        self.custom = {}
        """
        Custom carrier field enabling user to add parameters of their own.
        """

    def __str__(self):
        return "\n".join(f"{k.ljust(23)}= {v}" for k, v in self.__dict__.items())

    def to_file(self, file: str) -> None:
        """
        Save the current configuration to a file

        :param file: The path name
        """
        with open(file, "w") as f:
            f.write(self.to_json())

    @staticmethod
    def from_file(file: str) -> 'Config':
        """
        Load a configuration from a file to a new instance of Config

        :param file: The path name
        :return: A fresh instance of Config
        """
        raw = Path(file).read_text()
        return Config.from_json(raw)

    @staticmethod
    def from_json(s: str) -> 'Config':
        """
        Load a configuration from a json input to a new instance of Config

        :param s: The JSON text
        :return: A fresh instance of Config
        """
        data = json.loads(s)
        c = Config()
        for k, v in data.items():
            if hasattr(c, k):
                mapping = {"coverage_strategy": CoverageStrategy, "smt_solver": SmtSolver,
                           "seed_format": SeedFormat}
                if k in mapping:
                    v = mapping[k][v]
                elif k == "branch_solving_strategy":
                    v = reduce(lambda acc, x: BranchSolvingStrategy[x] | acc, v, 0)
                setattr(c, k, v)
            else:
                logging.warning(f"config unknown parameter: {k}")
        return c

    def to_json(self) -> str:
        """
        Convert the current configuration to a json output

        :return: JSON text
        """
        def to_str_list(value):
            return [x.name for x in list(BranchSolvingStrategy) if x in value]
        d = {}
        for k, v in self.__dict__.items():
            if isinstance(v, IntFlag):
                d[k] = to_str_list(v)
            elif isinstance(v, Enum):
                d[k] = v.name
            else:
                d[k] = v
        return json.dumps(d, indent=2)

    def is_format_composite(self) -> bool:
        """ Return true if the seed format is composite"""
        return self.seed_format == SeedFormat.COMPOSITE

    def is_format_raw(self) -> bool:
        """ Return true if the seed format is raw """
        return self.seed_format == SeedFormat.RAW



from triton import TritonContext
import bisect
from typing import Optional, Union, Generator, List
from collections import namedtuple
import struct
from contextlib import contextmanager

from tritondse.types import Perm, Addr, ByteSize, Endian

MemMap = namedtuple('Map', "start size perm name")


class MapOverlapException(Exception):
    """
    Exception raised when trying to map a memory area where some of
    the addresses overlap with an already mapped area.
    """
    pass


class MemoryAccessViolation(Exception):
    """
    Exception triggered when accessing memory with
    the wrong permissions.
    """
    def __init__(self, addr: Addr, access: Perm, map_perm: Perm = None, memory_not_mapped: bool = False, perm_error: bool = False):
        """
        :param addr: address where the violation occured
        :param access: type of access performed
        :param map_perm: permission of the memory page of `address`
        :param memory_not_mapped: whether the address was mapped or not
        :param perm_error: whether it is a permission error
        """
        super(MemoryAccessViolation, self).__init__()
        self.address: Addr = addr
        """ address where the violation occurred"""
        self._is_mem_unmapped = memory_not_mapped
        self._is_perm_error = perm_error
        self.access: Perm = access
        """Access type that was performed"""
        self.map_perm: Optional[Perm] = map_perm
        """Permissions of the memory map associated to the address"""

    def is_permission_error(self) -> bool:
        """True if the exception was caused by a permission issue"""
        return self._is_perm_error

    def is_memory_unmapped_error(self) -> bool:
        """
        Return true if the exception was raised due to an access
        to an area not mapped
        """
        return self._is_mem_unmapped

    def __str__(self) -> str:
        if self.is_permission_error():
            return f"(addr:{self.address:#08x}, access:{str(self.access)} on map:{str(self.map_perm)})"
        else:
            return f"({str(self.access)}: {self.address:#08x} unmapped)"

    def __repr__(self):
        return str(self)


STRUCT_MAP = {
    (True, 1): 'B',
    (False, 1): 'b',
    (True, 2): 'H',
    (False, 2): 'h',
    (True, 4): 'I',
    (False, 4): 'i',
    (True, 8): 'Q',
    (False, 8): 'q'
}

ENDIAN_MAP = {
    Endian.LITTLE: "<",
    Endian.BIG: ">"
}


class Memory(object):
    """
    Memory representation of the current :py:class:`ProcessState` object.
    It wraps all interaction with Triton's memory context to provide high-level
    function. It adds a segmentation and memory permission model at the top
    of Triton. It also overrides __getitem__ and the slice mechanism to be able
    read and write concrete memory values in a Pythonic manner.
    """

    def __init__(self, ctx: TritonContext, endianness: Endian = Endian.LITTLE):
        """
        :param ctx: TritonContext to interface with
        """
        self.ctx: TritonContext = ctx
        """Underlying Triton context"""
        self._linear_map_addr = []  # List of [map_start, map_end, map_start, map_end ...]
        self._linear_map_map = []   # List of [MemMap,    None,    MemMap,    None    ...]
        self._segment_enabled = True
        self._endian = endianness
        self._endian_key = ENDIAN_MAP[self._endian]
        self._mem_cbs_enabled = True
        # self._maps = {}  # Addr: -> Map

    def set_endianess(self, en: Endian) -> None:
        """
        Set the endianness of memory accesses. By default,
        endianess is little.

        :param en: Endian: Endianess to use.
        :return: None
        """
        self._endian = en
        self._endian_key = ENDIAN_MAP[self._endian]

    @property
    def _ptr_size(self) -> int:
        return self.ctx.getGprSize()

    @property
    def segmentation_enabled(self) -> bool:
        """
        returns whether segmentation enforcing is enabled

        :return: True if segmentation is enabled
        """
        return self._segment_enabled

    def disable_segmentation(self) -> None:
        """
        Turn-off segmentation enforcing.
        """
        self._segment_enabled = False

    def enable_segmentation(self) -> None:
        """
        Turn-off segmentation enforcing.
        """
        self._segment_enabled = True

    def set_segmentation(self, enabled: bool) -> None:
        """
        Set the segmentation enforcing with the given boolean.
        """
        self._segment_enabled = enabled

    @contextmanager
    def without_segmentation(self, disable_callbacks=False) -> Generator['Memory', None, None]:
        """
        Context manager enabling manipulating temporarily the memory
        without considering the memory permissions.
        E.g: It enables writing data in a memory mapped in RX
        :param disable_callbacks: Whether to disable memory callbacks that could have been set
        :return:
        """
        previous = self._segment_enabled
        self.disable_segmentation()
        cbs = self._mem_cbs_enabled
        self._mem_cbs_enabled = not disable_callbacks
        yield self
        self._mem_cbs_enabled = cbs
        self.set_segmentation(previous)

    def callbacks_enabled(self) -> bool:
        """
        Return whether memory callbacks are enabled.

        :return: True if callbacks are enabled
        """
        return self._mem_cbs_enabled

    def get_maps(self) -> Generator[MemMap, None, None]:
        """
        Iterate all the memory maps defined, including all memory
        areas allocated on the heap.

        :return: generator of all :py:class:`MemMap` objects
        """
        yield from (x for x in self._linear_map_map if x)

    def map(self, start, size, perm: Perm = Perm.R | Perm.W | Perm.X, name="") -> MemMap:
        """
        Map the given address and size in memory with the given permission.

        :raise MapOverlapException: In the case the map overlap an existing mapping
        :param start: address to map
        :param size: size to map
        :param perm: permission
        :param name: name to given to the memory region
        :return: MemMap freshly mapped
        """
        def _map_idx(idx):
            self._linear_map_addr.insert(idx, start + size - 1)  # end address is included
            self._linear_map_addr.insert(idx, start)
            self._linear_map_map.insert(idx, None)
            memmap = MemMap(start, size, perm, name)
            self._linear_map_map.insert(idx, memmap)
            return memmap

        if not self._linear_map_addr:  # Nothing mapped yet
            return _map_idx(0)

        idx = bisect.bisect_left(self._linear_map_addr, start)

        if idx == len(self._linear_map_addr):  # It should be mapped at the end
            return _map_idx(idx)

        addr = self._linear_map_addr[idx]
        if (idx % 2) == 0:  # We are on a start address
            if start < addr and start+size <= addr:  # Can fit before
                return _map_idx(idx)
            else:  # there is an overlap
                raise MapOverlapException(f"0x{start:08x}:{size} overlap with map: 0x{addr:08x} (even)")
        else:  # We are on an end address
            prev = self._linear_map_addr[idx-1]
            raise MapOverlapException(f"0x{start:08x}:{size} overlap with map: 0x{prev:08x} (odd)")

    def unmap(self, addr: Addr) -> None:
        """
        Unmap the :py:class:`MemMap` object mapped at the address.
        The address can be within the map and not requires pointing
        at the head.

        :param addr: address to unmap
        :return: None
        """
        def _unmap_idx(idx):
            self._linear_map_addr.pop(idx) # Pop the start
            self._linear_map_addr.pop(idx) # Pop the end
            self._linear_map_map.pop(idx)  # Pop the object
            self._linear_map_map.pop(idx)  # Pop the None padding

        idx = bisect.bisect_left(self._linear_map_addr, addr)
        try:
            mapaddr = self._linear_map_addr[idx]
            if (idx % 2) == 0:  # We are on a start address (meaning we should be exactly on map start other unmapped)
                if addr == mapaddr:  # We are exactly on the map address
                    _unmap_idx(idx)
                else:
                    raise MemoryAccessViolation(addr, Perm(0), memory_not_mapped=True)
            else:  # We are on an end address
                _unmap_idx(idx-1)
        except IndexError:
            raise MemoryAccessViolation(addr, Perm(0), memory_not_mapped=True)

    def mprotect(self, addr: Addr, perm: Perm) -> None:
        """
        Update the map at the given address with permissions provided in argument.

        :param addr: address of the map of which to change permission
        :param perm: permission to assign
        :return: None
        """
        idx = bisect.bisect_left(self._linear_map_addr, addr)
        try:
            if (idx % 2) == 0:  # We are on a start address (meaning we should be exactly on map start other unmapped)
                map = self._linear_map_map[idx]
                self._linear_map_map[idx] = MemMap(map.start, map.size, perm, map.name)  # replace map with new perms
            else:  # We are on an end address
                map = self._linear_map_map[idx-1]
                self._linear_map_map[idx-1] = MemMap(map.start, map.size, perm, map.name)  # replace map with new perms
        except IndexError:
            raise MemoryAccessViolation(addr, Perm(0), memory_not_mapped=True)

    def __setitem__(self, key: Addr, value: bytes) -> None:
        """
        Assign the given value at the address given by the key.
        The value must be bytes but can be multiple bytes.
        Warning: You cannot use the slice API on this function.

        :param key: address to write to
        :param value: content to write
        :raise MemoryAccessViolation: in case of invalid access
        """
        if isinstance(key, slice):
            raise TypeError("slice unsupported for __setitem__")
        else:
            self.write(key, value)

    def __getitem__(self, item: Union[Addr, slice]) -> bytes:
        """
        Read the memory at the given address. If the key
        is an integer reads a single byte. If the key is
        a slice: read addr+size bytes in memory.

        :param item: address, or address:size to read
        :return: memory content
        :raise MemoryAccessViolation: if the access is invalid
        """
        if isinstance(item, slice):
            return self.read(item.start, item.stop)
        elif isinstance(item, int):
            return self.read(item, 1)

    def write(self, addr: Addr, data: bytes) -> None:
        """
        Write the given `data` bytes at `addr` address.

        :param addr: address where to write
        :param data: data to write
        :return: None
        """
        if self._segment_enabled:
            map = self._get_map(addr, len(data))
            if map is None:
                raise MemoryAccessViolation(addr, Perm.W, memory_not_mapped=True)
            if Perm.W not in map.perm:
                raise MemoryAccessViolation(addr, Perm.W, map_perm=map.perm, perm_error=True)
        return self.ctx.setConcreteMemoryAreaValue(addr, data)

    def read(self, addr: Addr, size: ByteSize) -> bytes:
        """
        Read `size` bytes at `addr` address.

        :param addr: address to read
        :param size: size of content to read
        :return: bytes read
        """
        if self._segment_enabled:
            map = self._get_map(addr, size)
            if map is None:
                raise MemoryAccessViolation(addr, Perm.R, memory_not_mapped=True)
            if Perm.R not in map.perm:
                raise MemoryAccessViolation(addr, Perm.R, map_perm=map.perm, perm_error=True)
        return self.ctx.getConcreteMemoryAreaValue(addr, size)

    def _get_map(self, ptr: Addr, size: ByteSize) -> Optional[MemMap]:
        """
        Internal function returning the MemMap object associated
        with any address. It returns None if part of the memory
        range falls out of a memory mapping.
        Complexity is O(log(n))

        :param ptr: address in memory
        :param size: size of the memory
        :return: True if mapped
        """
        idx = bisect.bisect_left(self._linear_map_addr, ptr)
        try:
            addr = self._linear_map_addr[idx]
            if (idx % 2) == 0:  # We are on a start address (meaning we should be exactly on map start other unmapped)
                end = self._linear_map_addr[idx+1]
                return self._linear_map_map[idx] if (ptr == addr and ptr+size <= end+1) else None
            else:  # We are on an end address
                start = self._linear_map_addr[idx-1]
                return self._linear_map_map[idx-1] if (start <= addr and ptr+size <= addr+1) else None  # fit into the map
        except IndexError:
            return None  # Either raised when linear_map is empty or the address is beyond everything that is mapped

    def get_map(self, addr: Addr, size: ByteSize = 1) -> Optional[MemMap]:
        """
        Find the MemMap associated with the given address and returns
        it if any.

        :param addr: Address of the map (or any map inside)
        :param size: size of bytes for which we want the map
        :return: MemMap if found
        """
        return self._get_map(addr, size)

    def find_map(self, name: str) -> Optional[List[MemMap]]:
        """
        Find a map given its name.

        :param name: Map name
        :return: MemMap if found
        """
        l = []
        for map in (x for x in self._linear_map_map if x):
            if map.name == name:
                l.append(map)
        return l

    def map_from_name(self, name: str) -> MemMap:
        """
        Return a map from its name. This function assumes
        the map is present.

        :raise AssertionError: If the map is not found
        :param name: Map name
        :return: MemMap
        """
        for map in (x for x in self._linear_map_map if x):
            if map.name == name:
                return map
        assert False

    def is_mapped(self, ptr: Addr, size: ByteSize = 1) -> bool:
        """
        The function checks whether the memory is mapped or not.
        The implementation return False if the memory chunk overlap
        on two memory regions.
        Complexity is O(log(n))

        :param ptr: address in memory
        :param size: size of the memory
        :return: True if mapped
        """
        return self._get_map(ptr, size) is not None

    def has_ever_been_written(self, ptr: Addr, size: ByteSize) -> bool:
        """
        Returns whether the given range of addresses has previously
        been written or not. (Do not take in account the memory mapping).

        :param ptr: The pointer to check
        :type ptr: :py:obj:`tritondse.types.Addr`
        :param size: Size of the memory range to check
        :return: True if all addresses have been defined
        """
        return self.ctx.isConcreteMemoryValueDefined(ptr, size)

    def read_uint(self, addr: Addr, size: ByteSize = 4):
        """
        Read in the process memory a **little-endian** integer of the ``size`` at ``addr``.

        :param addr: Address at which to read data
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: Number of bytes to read
        :type size: Union[str, :py:obj:`tritondse.types.ByteSize`]
        :return: Integer value read
        :raise struct.error: If value can't fit in `size`
        """
        data = self.read(addr, size)
        return struct.unpack(self._endian_key+STRUCT_MAP[(True, size)], data)[0]

    def read_sint(self, addr: Addr, size: ByteSize = 4):
        """
        Read in the process memory a **little-endian** integer of the ``size`` at ``addr``.

        :param addr: Address at which to read data
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: Number of bytes to read
        :type size: Union[str, :py:obj:`tritondse.types.ByteSize`]
        :return: Integer value read
        :raise struct.error: If value can't fit in `size`
        """
        data = self.read(addr, size)
        return struct.unpack(self._endian_key+STRUCT_MAP[(False, size)], data)[0]

    def read_ptr(self, addr: Addr) -> int:
        """
        Read in the process memory a little-endian integer of size :py:attr:`tritondse.ProcessState.ptr_size`

        :param addr: Address at which to read data
        :type addr: :py:obj:`tritondse.types.Addr`
        :return: Integer value read
        """
        return self.read_uint(addr, self._ptr_size)

    def read_char(self, addr: Addr) -> int:
        """
        Read a char in memory (1-byte) following endianess.

        :param addr: address to read
        :return: char value as int
        """
        return self.read_sint(addr, 1)

    def read_uchar(self, addr: Addr) -> int:
        """
        Read an unsigned char in memory (1-byte) following endianness.

        :param addr: address to read
        :return: unsigned char value as int
        """
        return self.read_uint(addr, 1)

    def read_int(self, addr: Addr) -> int:
        """
        Read a signed integer in memory (4-byte) following endianness.

        :param addr: address to read
        :return: signed integer value as int
        """
        return self.read_sint(addr, 4)

    def read_word(self, addr: Addr) -> int:
        """
        Read signed word in memory (2-byte) following endianness.

        :param addr: address to read
        :return: signed word value as int
        """
        return self.read_uint(addr, 2)

    def read_dword(self, addr: Addr) -> int:
        """
        Read signed double word in memory (4-byte) following endianness.

        :param addr: address to read
        :return: dword value as int
        """
        return self.read_uint(addr, 4)

    def read_qword(self, addr: Addr) -> int:
        """
        Read signed qword in memory (8-byte) following endianness.

        :param addr: address to read
        :return: qword value as int
        """
        return self.read_uint(addr, 8)

    def read_long(self, addr: Addr) -> int:
        """
        Read 'C style' long in memory (4-byte) following endianness.

        :param addr: address to read
        :return: value as int
        """
        return self.read_sint(addr, 4)

    def read_ulong(self, addr: Addr) -> int:
        """
        Read unsigned long in memory (4-byte) following endianness.

        :param addr: address to read
        :return: unsigned long value as int
        """
        return self.read_uint(addr, 4)

    def read_long_long(self, addr: Addr) -> int:
        """
        Read long long in memory (8-byte) following endianness.

        :param addr: address to read
        :return: long long value as int
        """
        return self.read_sint(addr, 8)

    def read_ulong_long(self, addr: Addr) -> int:
        """
        Read unsigned long long in memory (8-byte) following endianness.

        :param addr: address to read
        :return: unsigned long long value as int
        """
        return self.read_uint(addr, 8)

    def read_string(self, addr: Addr) -> str:
        """ Read a string in process memory at the given address

        .. warning:: The memory read is unbounded. Thus the memory is iterated up until
                     finding a 0x0.

        :returns: the string read in memory
        :rtype: str
        """
        s = ""
        index = 0
        while True:
            val = self.read_uint(addr+index, 1)
            if not val:
                return s
            s += chr(val)
            index += 1

    def write_int(self, addr: Addr, value: int, size: ByteSize = 4):
        """
        Write in the process memory the given integer value of the given size at
        a specific address.

        :param addr: Address at which to read data
        :param value: data to write represented as an integer
        :param size: Number of bytes to read
        :raise struct.error: If integer value cannot fit in `size`
        """
        self.write(addr, struct.pack(self._endian_key+STRUCT_MAP[(value >= 0, size)], value))

    def write_ptr(self, addr: Addr, value: int) -> None:
        """
        Similar to :py:meth:`write_int` but the size is automatically adjusted
        to be ``ptr_size``.

        :param addr: address where to write data
        :type addr: :py:obj:`tritondse.types.Addr`
        :param value: pointer value to write
        :type value: int
        :raise struct.error: If integer value cannot fit in a pointer size
        """
        self.write_int(addr, value, self._ptr_size)

    def write_char(self, addr: Addr, value: int) -> None:
        """
        Write the integer value as a single byte in memory.

        :param addr: address to write
        :param value: integer value
        :raise struct.error: If integer value do not fit in a byte (>255)
        """
        self.write_int(addr, value, 1)

    def write_word(self, addr: Addr, value: int) -> None:
        """
        Write the word (2-byte) in memory following endianess.

        :param addr: address to write
        :param value: integer value
        :raise struct.error: If integer value do not fit in a word
        """
        self.write_int(addr, value, 2)

    def write_dword(self, addr: Addr, value: int) -> None:
        """
        Write the word (4-byte) in memory following endianess.

        :param addr: address to write
        :param value: integer value
        :raise struct.error: If integer value do not fit in a dword
        """
        self.write_int(addr, value, 4)

    def write_qword(self, addr: Addr, value: int) -> None:
        """
        Write the qword (8-byte) in memory following endianess.

        :param addr: address to write
        :param value: integer value
        :raise struct.error: If integer value do not fit in a qword
        """
        self.write_int(addr, value, 8)

    def write_long(self, addr: Addr, value: int) -> None:
        """
        Write a "C style" long (4-byte) in memory following endianess.

        :param addr: address to write
        :param value: integer value
        :raise struct.error: If integer value do not fit in a long
        """
        return self.write_int(addr, value, 4)

    def write_long_long(self, addr: Addr, value: int) -> None:
        """
        Write the "C style" long long (8-byte) in memory following endianess.

        :param addr: address to write
        :param value: integer value
        :raise struct.error: If integer value do not fit in a long long
        """
        return self.write_int(addr, value, 8)



from tritondse import ProbeInterface, CbType, SymbolicExecutor, ProcessState, SymbolicExplorator
import tritondse.logging

logger = tritondse.logging.get("probe.basictrace")


class BasicDebugTrace(ProbeInterface):
    """
    Basic probe that print instruction trace
    to logging debug.
    """
    NAME = "debugtrace-probe"

    def __init__(self):
        super(BasicDebugTrace, self).__init__()
        self._add_callback(CbType.PRE_INST, self.trace_debug)

    def trace_debug(self, exec: SymbolicExecutor, __: ProcessState, ins: 'Instruction'):
        logger.debug(f"[tid:{ins.getThreadId()}] {exec.trace_offset} [0x{ins.getAddress():x}]: {ins.getDisassembly()}")



class BasicTextTrace(ProbeInterface):
    """
    Basic probe that generate a txt execution trace
    for each run.
    """
    NAME = "txttrace-probe"

    def __init__(self):
        super(BasicTextTrace, self).__init__()
        self._add_callback(CbType.PRE_EXEC, self.pre_execution)
        self._add_callback(CbType.POST_EXEC, self.post_execution)
        self._add_callback(CbType.PRE_INST, self.trace_debug)

        # File in which to write the trace
        self._file = None

    def pre_execution(self, executor: SymbolicExecutor, _: ProcessState):
        # Triggered before each execution
        name = f"{executor.uid}-{executor.seed.hash}.txt"
        file = executor.workspace.get_metadata_file_path(f"{self.NAME}/{name}")
        self._file = open(file, "w")

    def post_execution(self, executor: SymbolicExecutor, _: ProcessState):
        self._file.close()

    def trace_debug(self, exec: SymbolicExecutor, __: ProcessState, ins: 'Instruction'):
        """
        This function is mainly used for debug.

        :param _: The current symbolic executor
        :param __: The current processus state of the execution
        :param ins: The current instruction executed
        :return: None
        """
        self._file.write(f"[tid:{ins.getThreadId()}] {exec.trace_offset} [0x{ins.getAddress():x}]: {ins.getDisassembly()}\n")



from enum import Enum, auto
import enum_tools.documentation
from triton import TritonContext


@enum_tools.documentation.document_enum
class ThreadState(Enum):
    RUNNING = auto()  # doc: Normal state
    DEAD = auto()     # doc: State after pthread_exit
    JOINING = auto()  # doc: State after pthread_join
    LOCKED = auto()   # doc: State after a pthread_lock & co


class ThreadContext(object):
    """
    Thread data structure holding all information related to it.
    Purposely used to save registers and to restore them in a
    TritonContext.
    """

    def __init__(self, tid: int):
        """
        :param tid: thread id
        """
        self.cregs = dict()          # context of concrete registers
        self.sregs = dict()          # context of symbolic registers
        self._join_th_id = None      # joined thread id
        self.tid = tid               # the thread id
        self.count = 0               # Number of instructions executed until scheduling
        self.state = ThreadState.RUNNING
        # FIXME: Keep the thread_scheduling and automated the reset on restore

    def save(self, tt_ctx: TritonContext) -> None:
        """
        Save the current thread state from the current execution.
        That implies keeping a reference on symbolic and concrete
        registers.

        :param tt_ctx: current TritonContext to save
        :type tt_ctx: `TritonContext <https://triton.quarkslab.com/documentation/doxygen/py_TritonContext_page.html>`_
        """
        # Save symbolic registers
        self.sregs = tt_ctx.getSymbolicRegisters()
        # Save concrete registers
        for r in tt_ctx.getParentRegisters():
            self.cregs.update({r.getId(): tt_ctx.getConcreteRegisterValue(r)})

    def restore(self, tt_ctx: TritonContext) -> None:
        """
        Restore a thread state in the given TritonContext

        :param tt_ctx: context in which to restor the current thread state
        :type tt_ctx: `TritonContext <https://triton.quarkslab.com/documentation/doxygen/py_TritonContext_page.html>`_
        """
        # Restore concrete registers
        for rid, v in self.cregs.items():
            tt_ctx.setConcreteRegisterValue(tt_ctx.getRegister(rid), v)
        # Restore symbolic registers
        for rid, e in self.sregs.items():
            tt_ctx.assignSymbolicExpressionToRegister(e, tt_ctx.getRegister(rid))

    def kill(self) -> None:
        """
        Kill the current thread. Called when exiting the thread.

        :return:
        """
        self.state = ThreadState.DEAD

    def is_dead(self) -> bool:
        """
        Returns whether the thread is killed or not

        :return: boolean indicating if the thread is dead or not
        """
        return self.state == ThreadState.DEAD

    def join_thread(self, th_id: int) -> None:
        """
        Put the thread in a join state where waits for
        another thread.

        :param th_id: id of the thread to join
        :return: None
        """
        self._join_th_id = th_id
        self.state = ThreadState.JOINING

    def is_waiting_to_join(self) -> bool:
        """
        Checks whether the thread is waiting to join
        another one.

        :return: boolean on whether it waits for another thread
        """
        return self.state == ThreadState.JOINING

    def cancel_join(self) -> None:
        """
        Cancel a join operation.

        :return: None
        """
        self._join_th_id = None
        self.state = ThreadState.RUNNING

    def is_main_thread(self) -> bool:
        """
        Returns whether or not it is the main thread
        (namely its id is 0)

        :return: bool
        """
        return self.tid == 0

    def is_running(self) -> bool:
        """
        Return if the thread is properly running or not.

        :return: True if the thread is running
        """
        return self.state == ThreadState.RUNNING



# built-ins
from __future__ import annotations

import io
import struct
import sys
import time
from typing import Union, Callable, Tuple, Optional, List, Dict


# third-party
# import z3  # For direct value enumeration
from triton import TritonContext, MemoryAccess, CALLBACK, CPUSIZE, Instruction, MODE, AST_NODE, SOLVER, EXCEPTION

# local imports
from tritondse.thread_context import ThreadContext
from tritondse.heap_allocator import HeapAllocator
from tritondse.types import Architecture, Addr, ByteSize, BitSize, PathConstraint, Register, Expression, \
                            AstNode, Registers, SolverStatus, Model, SymbolicVariable, ArchMode, Perm, FileDesc, Endian
from tritondse.arch import ARCHS, CpuState
from tritondse.loaders import Loader
from tritondse.memory import Memory, MemoryAccessViolation
import tritondse.logging

logger = tritondse.logging.get()


class ProcessState(object):
    """
    Current process state. This class keeps all the runtime related to a running
    process, namely current, instruction, thread, memory maps, file descriptors etc.
    It also wraps Triton execution and thus hold its context. At the top of this,
    it provides a user-friendly API to access data in both the concrete and symbolic
    state of Triton.
    """

    STACK_SEG = "[stack]"
    EXTERN_SEG = "[extern]"

    def __init__(self, endianness: Endian = Endian.LITTLE, time_inc_coefficient: float = 0.0001):
        """
        :param endianness: Endianness to consider
        :param time_inc_coefficient: Time coefficient to represent execution time of an
                                     instruction see: :py:attr:`tritondse.Config.time_inc_coefficient`
        """
        # EXTERN_BASE is a "fake" memory area (not mapped) that will
        # which addresses will be used for external symbols
        self.EXTERN_FUNC_BASE = 0x01000000  # Not PLT but a dummy address space containing pointers to external symbols

        # This range will be dynamically allocated
        # upon request.
        self.BASE_HEAP = 0x10000000
        self.END_HEAP = 0x6fffffff

        # The Triton's context
        self.tt_ctx = TritonContext()
        """TritonContext object"""
        self.actx: 'AstContext' = self.tt_ctx.getAstContext()
        """
        Triton `AstContext <https://triton-library.github.io/documentation/doxygen/py_AstContext_page.html>`_
        enabling crafting logical expressions to be solved by SMT
        """

        # Cpu object wrapping registers values
        self.cpu: CpuState = None  #: CpuState holding concrete values of registers *(initialized when calling load)*
        self._archinfo = None

        # Memory object
        self.memory: Memory = Memory(self.tt_ctx, endianness)
        """Memory object associated with the ProcessState """

        # Used to define that the process must exist
        self.stop = False

        # Signals table used by raise(), signal(), etc.
        #self.signals_table = dict()

        # Dynamic symbols name -> addr (where they are mapped)
        self.dynamic_symbol_table: Dict[str, Tuple[Addr, bool]] = {}
        """Dictionnary of dynamic symbols as retrieved during the loading"""

        # File descriptors table used by fopen(), fprintf(), etc.
        self._fd_table = {
            0: FileDesc(0, "stdin", sys.stdin),
            1: FileDesc(1, "stdout", sys.stdout),
            2: FileDesc(2, "stderr", sys.stderr),
        }
        # Unique file id incrementation
        self._fd_id = len(self._fd_table)

        # Allocation information used by malloc()
        self.heap_allocator: HeapAllocator = HeapAllocator(self.BASE_HEAP, self.END_HEAP, self.memory)
        """Allocator providing alloc, free primitives atop the Memory object"""

        # Unique thread id incrementation
        self._utid = 0

        # Current thread id
        self._tid = self._utid

        # Threads contexts
        self._threads = {
            self._tid: ThreadContext(self._tid)
        }

        # Thread mutext init magic number
        self.PTHREAD_MUTEX_INIT_MAGIC = 0xdead

        # Mutex and semaphore
        self.mutex_locked = False
        self.semaphore_locked = False

        # The time when the ProcessState is instancied.
        # It's used to provide a deterministic behavior when calling functions
        # like gettimeofday(), clock_gettime(), etc.
        self.time = time.time()

        # Configuration values
        self.endianness = endianness  #: Current endianness
        self.time_inc_coefficient = time_inc_coefficient

        # Runtime temporary variables
        self.__pcs_updated = False

        # The current instruction executed
        self.__current_inst = None

        # The memory mapping of the program ({vaddr_s : vaddr_e})
        self.__program_segments_mapping = {}

    @property
    def threads(self) -> List[ThreadContext]:
        """
        Gives a list of all threads currently active.

        :return:
        """
        return list(self._threads.values())

    @property
    def current_thread(self) -> ThreadContext:
        """
        Gives the current thread selected.

        :return: current thread
        :rtype: ThreadContext
        """
        return self._threads[self._tid]

    def switch_thread(self, thread: ThreadContext) -> bool:
        """
        Change the current thread to the one given in parameter.
        Thus save the current context, and restore the one of the
        thread given in parameter. It also resets the counter of
        the thread restored. If the current_thread is dead, it will
        also remove it !

        :param thread: thread to restore ThreadContext
        :return: True if the switch worked fine
        """
        assert (thread.tid in self._threads)

        try:
            if self.current_thread.is_dead():
                del self._threads[self._tid]
                # TODO: Finding all other threads joining it / (or locked by it ?) to unlock them
            else:  # Do a normal switch
                # Reset the counter and save its context
                self.current_thread.save(self.tt_ctx)

            # Schedule to the next thread
            thread.count = 0  # Reset the counter
            thread.restore(self.tt_ctx)
            self._tid = thread.tid
            return True

        except Exception as e:
            logger.error(f"Error while doing context switch: {e}")
            return False

    def spawn_new_thread(self, new_pc: Addr, args: Addr) -> ThreadContext:
        """
        Create a new thread in the process state. Parameters are the
        new program counter and a pointer to arguments to provide the thread.

        :param new_pc: new program counter (function to execution)
        :param args: arguments
        :return: thread context newly created
        """
        tid = self._get_unique_thread_id()
        thread = ThreadContext(tid)
        thread.save(self.tt_ctx)

        # Concretize pc, bp, sp, and first (argument)
        regs = [self.program_counter_register, self.stack_pointer_register, self.base_pointer_register, self._get_argument_register(0)]
        for reg in regs:
            if reg.getId() in thread.sregs:
                del thread.sregs[reg.getId()]

        thread.cregs[self.program_counter_register.getId()] = new_pc  # set new pc
        thread.cregs[self._get_argument_register(0).getId()] = args   # set args pointer
        stack = self.memory.map_from_name(self.STACK_SEG)
        thread.cregs[self.base_pointer_register.getId()] = ((stack.start+stack.size) - ((1 << 28) * tid))
        thread.cregs[self.stack_pointer_register.getId()] = ((stack.start+stack.size) - ((1 << 28) * tid))

        # Add the thread in the pool of threads
        self._threads[tid] = thread
        return thread

    def set_triton_mode(self, mode: MODE, value: int = True) -> None:
        """
        Set the given mode in the TritonContext.

        :param mode: mode to set in triton context
        :param value: value to set (default True)
        """
        self.tt_ctx.setMode(mode, value)

    def set_thumb(self, enable: bool) -> None:
        """
        Set thumb mode activated in the TritonContext. The mode will automatically
        be switched during execution, but at initialization this method enable
        activating it / disabling it. (Disabled be default)

        :param enable: bool: Wether or not to active thumb
        """
        self.tt_ctx.setThumb(enable)

    def set_solver_timeout(self, timeout: int) -> None:
        """
        Set the timeout for all subsequent queries.

        :param timeout: timeout in milliseconds
        """
        self.tt_ctx.setSolverTimeout(timeout)

    def set_solver(self, solver: Union[str, SOLVER]) -> None:
        """
        Set the SMT solver to use in the background.

        :param solver: Solver to use
        """
        if isinstance(solver, str):
            solver = getattr(SOLVER, solver.upper(), SOLVER.Z3)
        self.tt_ctx.setSolver(solver)

    def _get_unique_thread_id(self) -> int:
        """
        Return a new unique thread id. Used by thread related functions when spawning a new thread.

        :returns: new thread identifier
        """
        self._utid += 1
        return self._utid

    def create_file_descriptor(self, name: str, file: io.IOBase) -> FileDesc:
        """
        Create a new file descriptor out of a name.

        :param name: name of the file
        :param file: object to read from
        :return: FileDesc object
        """
        new_fd_id = self._fd_id
        self._fd_id += 1
        filedesc = FileDesc(id=new_fd_id, name=name, fd=file)
        self._fd_table[new_fd_id] = filedesc
        return filedesc

    def close_file_descriptor(self, id: int) -> None:
        """
        Close the given file descriptor id.

        :param id: id of the file descriptor
        :return: None
        """
        filedesc = self._fd_table.pop(id)
        if isinstance(filedesc.fd, io.IOBase):
            filedesc.fd.close()

    def get_file_descriptor(self, id: int) -> FileDesc:
        """
        Get the given file descriptor.

        :raise KeyError: if the file descriptor is not found
        :param id: id of the file descriptor
        :return: FileDesc object
        """
        return self._fd_table[id]

    def file_descriptor_exists(self, id: int) -> bool:
        """
        Returns whether the file descriptor has been defined or not.

        :param id: id of the file descriptor
        :return: True if the id is found
        """
        return bool(id in self._fd_table)

    @property
    def architecture(self) -> Architecture:
        """
        Architecture of the current process state

        :return: Architecture set
        """
        return Architecture(self.tt_ctx.getArchitecture())

    @architecture.setter
    def architecture(self, arch: Architecture) -> None:
        """
        Set the architecture of the process state.
        Internal set it in the TritonContext

        :param arch: Architecture to set
        """
        self.tt_ctx.setArchitecture(arch)

    @property
    def ptr_size(self) -> ByteSize:
        """
        Size of a pointer in bytes

        :rtype: :py:obj:`tritondse.types.ByteSize`
        """
        return self.tt_ctx.getGprSize()

    @property
    def ptr_bit_size(self) -> BitSize:
        """
        Size of a pointer in bits

        :rtype: :py:obj:`tritondse.types.BitSize`
        """
        return self.tt_ctx.getGprBitSize()

    @property
    def minus_one(self) -> int:
        """
        Value -1 according to the architecture size (32 or 64 bits)

        :return: -1 as an unsigned Python integer
        """
        return (1 << self.ptr_bit_size) - 1

    @property
    def registers(self) -> Registers:
        """
        All registers according to the current architecture defined.
        The object returned is the TritonContext.register object.

        :rtype: :py:obj:`tritondse.types.Registers`
        """
        return self.tt_ctx.registers

    @property
    def return_register(self) -> Register:
        """
        Return the appropriate return register according to the arch.

        :rtype: :py:obj:`tritondse.types.Register`
        """
        return getattr(self.registers, self._archinfo.ret_reg)

    @property
    def program_counter_register(self) -> Register:
        """
        Return the appropriate pc register according to the arch.

        :rtype: :py:obj:`tritondse.types.Register`
        """
        return getattr(self.registers, self._archinfo.pc_reg)

    @property
    def base_pointer_register(self) -> Register:
        """
        Return the appropriate base pointer register according to the arch.

        :rtype: :py:obj:`tritondse.types.Register`
        """
        return getattr(self.registers, self._archinfo.bp_reg)

    @property
    def stack_pointer_register(self) -> Register:
        """
        Return the appropriate stack pointer register according to the arch.

        :rtype: :py:obj:`tritondse.types.Register`
        """
        return getattr(self.registers, self._archinfo.sp_reg)

    @property
    def _syscall_register(self) -> Register:
        """ Return the appropriate syscall id register according to the arch """
        return getattr(self.registers, self._archinfo.sys_reg)

    def _get_argument_register(self, i: int) -> Register:
        """
        Return the appropriate register according to the arch.

        :raise: IndexError If the index is out of arguments bound
        :return: Register
        """
        return getattr(self.registers, self._archinfo.reg_args[i])

    def initialize_context(self, arch: Architecture):
        """
        Initialize the context with the given architecture

        .. todo:: Protecting that function

        :param arch: The architecture to initialize
        :type arch: Architecture
        :return: None
        """
        self.architecture = arch
        self._archinfo = ARCHS[self.architecture]
        self.cpu = CpuState(self.tt_ctx, self._archinfo)

    def unpack_integer(self, data: bytes, size: int) -> int:
        """
        Unpack the given bytes into into integer value respecting
        size given and endianness.

        :param data: bytes data to unpack
        :param size: size in bits of data to unpack
        :return: integer value unpacked
        """
        s = "<" if self.endianness == Endian.LITTLE else ">"
        tab = {8: 'B', 16: 'H', 32: 'I', 64: 'Q'}
        s += tab[size]
        return struct.unpack(s, data)[0]

    def pack_integer(self, value: int, size: int) -> bytes:
        """
        Unpack the given bytes into into integer value respecting
        size given and endianness.

        :param data: bytes data to unpack
        :param size: size in bits of data to unpack
        :return: integer value packed as bytes
        """
        s = "<" if self.endianness == Endian.LITTLE else ">"
        tab = {8: 'B', 16: 'H', 32: 'I', 64: 'Q'}
        s += tab[size]
        return struct.pack(s, value)

    def read_register(self, register: Union[str, Register]) -> int:
        """
        Read the current concrete value of the given register.

        :param register: string of the register or Register object
        :type register: Union[str, :py:obj:`tritondse.types.Register`]
        :return: Integer value
        """
        reg = getattr(self.tt_ctx.registers, register) if isinstance(register, str) else register  # if str transform to reg
        return self.tt_ctx.getConcreteRegisterValue(reg)

    def write_register(self, register: Union[str, Register], value: int) -> None:
        """
        Read the current concrete value of the given register.

        :param register: string of the register or Register object
        :type register: Union[str, :py:obj:`tritondse.types.Register`]
        :param value: integer value to assign in the register
        :type value: int
        """
        reg = getattr(self.tt_ctx.registers, register) if isinstance(register, str) else register  # if str transform to reg
        return self.tt_ctx.setConcreteRegisterValue(reg, value)

    def register_triton_callback(self, cb_type: CALLBACK, callback: Callable) -> None:
        """
        Register the given ``callback`` as triton callback to hook memory/registers
        read/writes.

        :param cb_type: Callback enum type as defined by Triton
        :type cb_type: `CALLBACK <https://triton.quarkslab.com/documentation/doxygen/py_CALLBACK_page.html>`_
        :param callback: routines to call on the given event
        """
        self.tt_ctx.addCallback(cb_type, callback)

    def clear_triton_callbacks(self) -> None:
        """
        Remove all registered callbacks in triton.
        """
        self.tt_ctx.clearCallbacks()

    def is_heap_ptr(self, ptr: Addr) -> bool:
        """
        Check whether a given address is pointing in the heap area.

        :param ptr: Address to check
        :type ptr: :py:obj:`tritondse.types.Addr`
        :return: True if pointer points to the heap area *(allocated or not)*.
        """
        if self.BASE_HEAP <= ptr < self.END_HEAP:
            return True
        return False

    def is_syscall(self) -> bool:
        """
        Check whether the current instrution fetched is a syscall or not.
        """
        return bool(self.current_instruction.getType() in self._archinfo.syscall_inst)

    def fetch_instruction(self, address: Addr = None, set_as_current: bool = True, disable_callbacks: bool = True) -> Instruction:
        """
        Fetch the instruction at the given address. If no address
        is specified the current program counter one is used.

        :raise MemoryAccessViolation: If the instruction cannot be fetched in the memory.

        :param address: address where to get the instruction from
        :param set_as_current: set as the current instruction in the process state
        :param disable_callbacks: whether memory callbacks should be disabled to fetch memory bytes
        :return: instruction disassembled
        """
        if address is None:
            address = self.cpu.program_counter
        with self.memory.without_segmentation(disable_callbacks=disable_callbacks):
            data = self.memory.read(address, 16)
        i = Instruction(address, data)
        i.setThreadId(self.current_thread.tid)
        self.tt_ctx.disassembly(i)  # This needs to be done before using i.getSize()
                                    # otherwise, i.getSize() will always be 16

        if self.memory.segmentation_enabled:
            map = self.memory.get_map(address, i.getSize())
            if map is None:
                raise MemoryAccessViolation(address, Perm.X, memory_not_mapped=True)
            if Perm.X not in map.perm:  # Note: in this model we can execute code in non-readable pages
                raise MemoryAccessViolation(address, Perm.X, map_perm=map.perm, perm_error=True)
        if set_as_current:
            self.__current_inst = i
        return i

    def process_instruction(self, instruction: Instruction) -> bool:
        """
        Process the given triton instruction on this process state.

        :param instruction: Triton Instruction object
        :type instruction: `Instruction <https://triton.quarkslab.com/documentation/doxygen/py_Instruction_page.html>`_
        :return: True if the processing of the instruction succeeded (False otherwise)
        """
        self.__pcs_updated = False
        __len_pcs = self.tt_ctx.getPathPredicateSize()

        if not instruction.getDisassembly():  # If the insrtuction has not been disassembled
            self.tt_ctx.disassembly(instruction)

        self.__current_inst = instruction
        ret = self.tt_ctx.buildSemantics(instruction)

        # Simulate that the time of an executed instruction is time_inc_coefficient.
        # For example, if time_inc_coefficient is 0.0001, it means that an instruction
        # takes 100us to be executed. Used to provide a deterministic behavior when
        # calling time functions (e.g gettimeofday(), clock_gettime(), ...).
        self.time += self.time_inc_coefficient

        if self.tt_ctx.getPathPredicateSize() > __len_pcs:
            self.__pcs_updated = True

        return ret == EXCEPTION.NO_FAULT

    @property
    def path_predicate_size(self) -> int:
        """
        Get the size of the path predicate (conjonction
        of all branches and additionnals constraints added)

        :return: size of the predicate
        """
        return self.tt_ctx.getPathPredicateSize()

    def is_path_predicate_updated(self) -> bool:
        """ Return whether or not the path predicate has been updated """
        return self.__pcs_updated

    @property
    def last_branch_constraint(self) -> PathConstraint:
        """
        Return the last PathConstraint object added in the path predicate.
        Should be called after :py:meth:`is_path_predicate_updated`.

        :raise IndexError: if the path predicate is empty
        :return: the path constraint object as returned by Triton
        :rtype: `PathConstraint <https://triton.quarkslab.com/documentation/doxygen/py_PathConstraint_page.html>`_
        """
        return self.tt_ctx.getPathConstraints()[-1]

    @property
    def current_instruction(self) -> Optional[Instruction]:
        """
        The current instruction being executed. *(None if not set yet)*

        :rtype: Optional[`Instruction <https://triton.quarkslab.com/documentation/doxygen/py_Instruction_page.html>`_]
        """
        return self.__current_inst

    def is_register_symbolic(self, register: Union[str, Register]) -> bool:
        """
        Check whether the register is symbolic or not.

        :param register: register string, or Register object
        :type register: Union[str, :py:obj:`tritondse.types.Register`]
        :return: True if the register is symbolic
        """
        reg = getattr(self.tt_ctx.registers, register) if isinstance(register, str) else register
        return self.tt_ctx.getRegisterAst(reg).isSymbolized()

    def read_symbolic_register(self, register: Union[str, Register]) -> Expression:
        """
        Get the symbolic expression associated with the given register.

        :param register: register string, or Register object
        :type register: Union[str, :py:obj:`tritondse.types.Register`]
        :return: SymbolicExpression of the register as returned by Triton
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        reg = getattr(self.tt_ctx.registers, register) if isinstance(register, str) else register  # if str transform to reg
        sym_reg = self.tt_ctx.getSymbolicRegister(reg)

        if sym_reg is None or reg.getBitSize() != sym_reg.getAst().getBitvectorSize():
            return self.tt_ctx.newSymbolicExpression(self.tt_ctx.getRegisterAst(reg))
        else:
            return sym_reg

    def write_symbolic_register(self, register: Union[str, Register], expr: Union[AstNode, Expression], comment: str = "") -> None:
        """
        Assign the given symbolic expression to the register. The given expression can either be a SMT AST node
        or directly an Expression (SymbolicExpression).

        :param register: register identifier (str or Register)
        :type register: Union[str, :py:obj:`tritondse.types.Register`]
        :param expr: expression to assign (`AstNode <https://triton.quarkslab.com/documentation/doxygen/py_AstNode_page.html>`_
               or `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_)
        :param comment: Comment to add on the symbolic expression created
        :type comment: str
        """
        reg = getattr(self.tt_ctx.registers, register) if isinstance(register, str) else register  # if str transform to reg
        exp = expr if hasattr(expr, "getAst") else self.tt_ctx.newSymbolicExpression(expr, f"assign {reg.getName()}: {comment}")
        self.write_register(reg, exp.getAst().evaluate())  # Update concrete state to keep sync
        self.tt_ctx.assignSymbolicExpressionToRegister(exp, reg)

    def read_symbolic_memory_int(self, addr: Addr, size: ByteSize) -> Expression:
        """
        Return a new Symbolic Expression representing the whole memory range given in parameter.
        That function should not be used on big memory chunks.

        :param addr: Memory address
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: memory size in bytes
        :type size: :py:obj:`tritondse.types.ByteSize`
        :raise RuntimeError: If the size if not aligned
        :return: Symbolic Expression associated with the memory
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        if size == 1:
            return self.read_symbolic_memory_byte(addr)
        elif size in [2, 4, 8, 16, 32, 64]:
            ast = self.tt_ctx.getMemoryAst(MemoryAccess(addr, size))
            return self.tt_ctx.newSymbolicExpression(ast)
        else:
            raise RuntimeError("size should be aligned [1, 2, 4, 8, 16, 32, 64] (bytes)")

    def read_symbolic_memory_byte(self, addr: Addr) -> Expression:
        """
        Thin wrapper to retrieve the symbolic expression of a single bytes in memory.

        :param addr: Memory address
        :type addr: :py:obj:`tritondse.types.Addr`
        :return: Symbolic Expression associated with the memory
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        res = self.tt_ctx.getSymbolicMemory(addr)
        if res is None:
            return self.tt_ctx.newSymbolicExpression(self.tt_ctx.getMemoryAst(MemoryAccess(addr, 1)))
        else:
            return res

    def read_symbolic_memory_bytes(self, addr: Addr, size: ByteSize) -> Expression:
        """
        Return a new Symbolic Expression representing the whole memory range given in parameter.
        That function should not be used on big memory chunks.

        :param addr: Memory address
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: memory size in bytes
        :type size: :py:obj:`tritondse.types.ByteSize`
        :return: Symbolic Expression associated with the memory
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        if size == 1:
            return self.read_symbolic_memory_byte(addr)
        else:  # Need to create a per-byte expression with concat
            asts = [self.tt_ctx.getMemoryAst(MemoryAccess(addr+i, CPUSIZE.BYTE)) for i in range(size)]
            concat_expr = self.actx.concat(asts)
            return self.tt_ctx.newSymbolicExpression(concat_expr)

    def write_symbolic_memory_int(self, addr: Addr, size: ByteSize, expr: Union[AstNode, Expression]) -> None:
        """
        Assign the given symbolic expression representing an integer to the given address.
        That function should not be used on big memory chunks.

        :param addr: Memory address
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: memory size in bytes
        :type size: :py:obj:`tritondse.types.ByteSize`
        :param expr: expression to assign (`AstNode <https://triton.quarkslab.com/documentation/doxygen/py_AstNode_page.html>`_
                     or `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_)
        :raise RuntimeError: if the size if not aligned
        """
        expr = expr if hasattr(expr, "getAst") else self.tt_ctx.newSymbolicExpression(expr, f"assign memory")
        if size in [1, 2, 4, 8, 16, 32, 64]:
            self.tt_ctx.setConcreteMemoryValue(MemoryAccess(addr, size), expr.getAst().evaluate())  # To keep the concrete state synchronized
            self.tt_ctx.assignSymbolicExpressionToMemory(expr, MemoryAccess(addr, size))
        else:
            raise RuntimeError("size should be aligned [1, 2, 4, 8, 16, 32, 64] (bytes)")

    def write_symbolic_memory_byte(self, addr: Addr, expr: Union[AstNode, Expression]) -> None:
        """
        Set a single bytes symbolic at the given address

        .. NOTE: We purposefully not provide a way to assign in memory a symbolic expression of
           arbitrary size as it would imply doing many extract on the given expression. For buffer
           you should do it in a per-byte manner with this method.

        :param addr: Memory address
        :type addr: :py:obj:`tritondse.types.Addr`
        :param expr: byte expression to assign (`AstNode <https://triton.quarkslab.com/documentation/doxygen/py_AstNode_page.html>`_
                     or `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_)
        """
        expr = expr if hasattr(expr, "getAst") else self.tt_ctx.newSymbolicExpression(expr, f"assign memory")
        ast = expr.getAst()
        assert ast.getBitvectorSize() == 8
        self.tt_ctx.setConcreteMemoryValue(MemoryAccess(addr, CPUSIZE.BYTE), ast.evaluate())  # Keep concrete state synced
        self.tt_ctx.assignSymbolicExpressionToMemory(expr, MemoryAccess(addr, CPUSIZE.BYTE))

    def is_memory_symbolic(self, addr: Addr, size: ByteSize) -> bool:
        """
        Iterate the symbolic memory and returns whether or not at least one byte of the buffer
        is symbolic

        :param addr: Memory address
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: size of the memory range to check
        :type size: :py:obj:`tritondse.types.ByteSize`
        :return: True if at least one byte of the memory is symbolic, false otherwise
        """
        for i in range(addr, addr+size):
            if self.tt_ctx.isMemorySymbolized(MemoryAccess(i, 1)):
                return True
        return False

    def push_constraint(self, constraint: AstNode, comment: str = "") -> None:
        """
        Thin wrapper on the triton context underneath to add a path constraint.

        :param constraint: Constraint expression to add
        :type constraint: `AstNode <https://triton.quarkslab.com/documentation/doxygen/py_AstNode_page.html>`_
        :param comment: String comment to attach to the constraint
        :type comment: str
        """
        self.tt_ctx.pushPathConstraint(constraint, comment)

    def get_path_constraints(self) -> List[PathConstraint]:
        """
        Get the list of all path constraints set in the Triton context.

        :return: list of constraints
        """
        return self.tt_ctx.getPathConstraints()

    def concretize_register(self, register: Union[str, Register]) -> None:
        """
        Concretize the given register with its current concrete value.
        **This operation is sound** as it will also add a path constraint
        to enforce that the symbolic register value is equal to its concrete
        value.

        :param register: Register identifier (str or Register)
        :type register: Union[str, :py:obj:`tritondse.types.Register`]
        """
        reg = getattr(self.tt_ctx.registers, register) if isinstance(register, str) else register
        if self.tt_ctx.isRegisterSymbolized(reg):
            value = self.read_register(reg)
            self.push_constraint(self.read_symbolic_register(reg).getAst() == value)
        # Else do not even push the constraint

    def concretize_memory_bytes(self, addr: Addr, size: ByteSize) -> None:
        """
        Concretize the given memory with its current concrete value.
        **This operation is sound** and allows restraining the memory
        value to its constant value.

        :param addr: Address to concretize
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: Size of the integer to concretize
        :type size: :py:obj:`tritondse.types.ByteSize`
        """
        data = self.memory.read(addr, size)
        if self.is_memory_symbolic(addr, size):
            self.push_constraint(self.read_symbolic_memory_bytes(addr, size).getAst() == data)
        # else do not even push the constraint

    def concretize_memory_int(self, addr: Addr, size: ByteSize) -> None:
        """
        Concretize the given memory with its current concrete value.
        **This operation is sound** and allows restraining the memory
        value to its constant value.

        :param addr: Address to concretize
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: Size of the integer to concretize
        :type size: :py:obj:`tritondse.types.ByteSize`
        """
        value = self.memory.read_uint(addr, size)
        if self.tt_ctx.isMemorySymbolized(MemoryAccess(addr, size)):
            self.push_constraint(self.read_symbolic_memory_int(addr, size).getAst() == value)
        # else do not even push the constraint

    def concretize_argument(self, index: int) -> None:
        """
        Concretize the given function parameter following the calling convention
        of the architecture.

        :param index: Argument index
        :type index: int
        """
        try:
            self.concretize_register(self._get_argument_register(index))
        except IndexError:
            len_args = len(self._archinfo.reg_args)
            addr = self.cpu.stack_pointer + self.ptr_size + ((index-len_args) * self.ptr_size)  # Retrieve stack address
            self.concretize_memory_int(addr, self.ptr_size)                     # Concretize the value at this addr

    def write_argument_value(self, i: int, val: int) -> None:
        """
        Write the parameter index with the given value. It will take in account
        whether the argument is in a register or the stack.

        :param i: Ith argument of the function
        :param val: integer value of the parameter
        :return: None
        """
        try:
            return self.write_register(self._get_argument_register(i), val)
        except IndexError:
            len_args = len(self._archinfo.reg_args)
            return self.write_stack_value(i-len_args, val, offset=1)

    def get_argument_value(self, i: int) -> int:
        """
        Get the integer value of parameters following the call convention.
        The value originate either from a register or the stack depending
        on the ith argument requested.

        :param i: Ith argument of the function
        :type i: int
        :return: integer value of the parameter
        :rtype: int
        """
        try:
            return self.read_register(self._get_argument_register(i))
        except IndexError:
            len_args = len(self._archinfo.reg_args)
            return self.get_stack_value(i-len_args, offset=1)

    def get_argument_symbolic(self, i: int) -> Expression:
        """
        Return the symbolic expression associated with the given ith parameter.

        :param i: Ith function parameter
        :return: Symbolic expression associated
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        try:
            return self.read_symbolic_register(self._get_argument_register(i))
        except IndexError:
            len_args = len(self._archinfo.reg_args)
            addr = self.cpu.stack_pointer + ((i-len_args) * self.ptr_size)
            return self.read_symbolic_memory_int(addr, self.ptr_size)

    def get_full_argument(self, i: int) -> Tuple[int, Expression]:
        """
        Get both the concrete argument value along with its symbolic expression.

        :return: Tuple containing concrete value and symbolic expression
        """
        return self.get_argument_value(i), self.get_argument_symbolic(i)

    def get_string_argument(self, idx: int) -> str:
        """Read a string for which address is a function parameter.
        The function first get the argument value, and then dereference
        the string located at that address.

        :param idx: argument index
        :type idx: int
        :returns: memory string
        :rtype: str
        """
        return self.memory.read_string(self.get_argument_value(idx))

    def get_format_string(self, addr: Addr) -> str:
        """
        Returns a formatted string in Python format from a format string
        located in memory at ``addr``.

        :param addr: Address to concretize
        :type addr: :py:obj:`tritondse.types.Addr`
        :rtype: str
        """
        return self.memory.read_string(addr)                                             \
               .replace("%s", "{}").replace("%d", "{}").replace("%#02x", "{:#02x}")     \
               .replace("%#x", "{:#x}").replace("%x", "{:x}").replace("%02X", "{:02X}") \
               .replace("%c", "{:c}").replace("%02x", "{:02x}").replace("%ld", "{}")    \
               .replace("%*s", "").replace("%lX", "{:X}").replace("%08x", "{:08x}")     \
               .replace("%u", "{}").replace("%lu", "{}").replace("%zu", "{}")           \
               .replace("%02u", "{:02d}").replace("%03u", "{:03d}")                     \
               .replace("%03d", "{:03d}").replace("%p", "{:#x}").replace("%i", "{}")

    def get_format_arguments(self, fmt_addr: Addr, args: List[int]) -> List[Union[int, str]]:
        """
        Read the format string at ``fmt_addr``. For each format item
        which are strings, dereference that associated string and replaces it
        in ``args``.

        :param fmt_addr: Address to concretize
        :type fmt_addr: :py:obj:`tritondse.types.Addr`
        :param args: Parameters associated with the format string
        :type args: List[int]
        :rtype: List[Union[int, str]]
        """
        # FIXME: Modifies inplace args (which is not very nice)
        s_str = self.memory.read_string(fmt_addr)
        postString = [i for i, x in enumerate([i for i, c in enumerate(s_str) if c == '%']) if s_str[x+1] == "s"]
        for p in postString:
            args[p] = self.memory.read_string(args[p])
            args[p] = args[p].encode("latin-1").decode()
        return args

    def get_stack_value(self, index: int, offset: int = 0) -> int:
        """
        Returns the value at the ith position further in the stack

        :param index: The index position from the top of the stack
        :type index: int
        :param offset: An integer value offset to apply to stack address
        :type offset: int
        :return: the value got
        :return: the value got
        :rtype: int
        """
        addr = self.cpu.stack_pointer + (offset * self.ptr_size) + (index * self.ptr_size)
        return self.memory.read_uint(addr, self.ptr_size)

    def write_stack_value(self, index: int, value: int, offset: int = 0) -> None:
        """
        Write the given value on the stack at the given index relative to the current
        stack pointer. The index value can be positive to write further down the stack
        or negative to write upward.

        :param index: The index position from the top of the stack
        :type index: int
        :param value: Integer value to write on the stack
        :type value: int
        :param offset: Add an optional ith item offset to add to stack value (not a size)
        :type offset: int
        :return: the value got
        :rtype: int
        """
        addr = self.cpu.stack_pointer + (offset * self.ptr_size) + (index * self.ptr_size)
        self.memory.write_int(addr, value, self.ptr_size)

    def pop_stack_value(self) -> int:
        """
        Pop a stack value, and the re-increment the stack pointer value.
        This operation is fully concrete.

        :return: int
        """
        val = self.memory.read_ptr(self.cpu.stack_pointer)
        self.cpu.stack_pointer += self.ptr_size
        return val

    def push_stack_value(self, value: int) -> None:
        """
        Push a stack value. It then decreement the stack pointer value.

        :param value: The value to push
        """
        self.memory.write_ptr(self.cpu.stack_pointer-self.ptr_size, value)
        self.cpu.stack_pointer -= self.ptr_size

    def is_halt_instruction(self) -> bool:
        """
        Check if the the current instruction is corresponding to an 'halt' instruction
        in the target architecture.

        :returns: Return true if on halt instruction architecture independent
        """
        halt_opc = self._archinfo.halt_inst
        return self.__current_inst.getType() == halt_opc

    def solve(self, constraint: Union[AstNode, List[AstNode]], with_pp: bool = True) -> Tuple[SolverStatus, Model]:
        """
        Solve the given constraint one the current symbolic state and returns both
        a Solver status and a model. If not SAT the model returned is empty. Argument
        ``with_pp`` enables checking the constraint taking in account the path predicate.

        :param constraint: AstNode or list of AstNodes constraints to solve
        :param with_pp: whether to take in account path predicate
        :return: tuple of status and model
        """
        if with_pp:
            cst = constraint if isinstance(constraint, list) else [constraint]
            final_cst = self.actx.land([self.tt_ctx.getPathPredicate()]+cst)
        else:
            final_cst = self.actx.land(constraint) if isinstance(constraint, list) else constraint

        model, status, _ = self.tt_ctx.getModel(final_cst, status=True)
        return SolverStatus(status), model

    def solve_no_pp(self, constraint: Union[AstNode, List[AstNode]]) -> Tuple[SolverStatus, Model]:
        """
        Helper function that solve a constraint forcing not to use
        the path predicate.

        .. warning:: Solving a query without the path predicate gives theoretically
                     unsound results.

        :param constraint: AstNode constraint to solve
        :return: tuple of status and model
        """
        return self.solve(constraint, with_pp=False)

    def symbolize_register(self, register: Union[str, Register], alias: str = None) -> SymbolicVariable:
        """
        Symbolize the given register. This a proxy for the symbolizeRegister
        Triton function.

        :param register: string of the register or Register object
        :type register: Union[str, :py:obj:`tritondse.types.Register`]
        :param alias: alias name to give to the symbolic variable
        :type alias: str
        :return: Triton Symbolic variable created
        """
        reg = getattr(self.tt_ctx.registers, register) if isinstance(register, str) else register  # if str get reg
        if alias:
            var = self.tt_ctx.symbolizeRegister(reg, alias)
        else:
            var = self.tt_ctx.symbolizeRegister(reg)
        return var

    def symbolize_memory_byte(self, addr: Addr, alias: str = None) -> SymbolicVariable:
        """
        Symbolize the given memory cell. Returns the associated
        SymbolicVariable

        :param addr: Address to symbolize
        :type addr: :py:obj:`tritondse.types.Addr`
        :param alias: alias to give the variable
        :return: newly created symbolic variable
        :rtype: :py:obj:`tritondse.types.SymbolicVariable`
        """
        if alias:
            return self.tt_ctx.symbolizeMemory(MemoryAccess(addr, CPUSIZE.BYTE), alias)
        else:
            return self.tt_ctx.symbolizeMemory(MemoryAccess(addr, CPUSIZE.BYTE))

    def symbolize_memory_bytes(self, addr: Addr, size: ByteSize, alias_prefix: str = None, offset: int = 0) -> List[SymbolicVariable]:
        """
        Symbolize a range of memory addresses. Can optionally provide an alias
        prefix.

        :param addr: Address at which to read data
        :type addr: :py:obj:`tritondse.types.Addr`
        :param size: Number of bytes to symbolize
        :type size: :py:obj:`tritondse.types.ByteSize`
        :param alias_prefix: prefix name to give the variable
        :type alias_prefix: str
        :return: list of Symbolic variables created
        :rtype: List[:py:obj:`tritondse.types.SymbolicVariable`]
        """
        if alias_prefix:
            return [self.symbolize_memory_byte(addr+i, alias_prefix+f"[{i+offset}]") for i in range(size)]
        else:
            return [self.symbolize_memory_byte(addr+i) for i in range(size)]

    def get_expression_variable_values_model(self, exp: Union[AstNode, Expression], model: Model) -> Dict[SymbolicVariable: int]:
        """
        Given a symbolic expression and a model, returns the valuation
        of all variables involved in the expression.

        :param exp: Symbolic Expression to look into
        :param model: Model generated by the solver
        :return: dictionnary of symbolic variables and their associated value (as int)
        """
        ast = exp.getAst() if hasattr(exp, "getAst") else exp
        ast_vars = self.actx.search(ast, AST_NODE.VARIABLE)
        sym_vars = [x.getSymbolicVariable() for x in ast_vars]
        final_dict = {}
        for avar, svar in zip(ast_vars, sym_vars):
            if svar.getId() in model:
                final_dict[svar] = model[svar.getId()].getValue()
            else:
                final_dict[svar] = avar.evaluate()
        return final_dict

    def evaluate_expression_model(self, exp: Union[AstNode, Expression], model: Model) -> int:
        """
        Evaluate the given expression on the given model. The value returned is the
        integer value corresponding to the bitvector evaluation of the expression.

        :param exp: Symbolic Expression to evaluate
        :param model: Model generated by the solver
        :return: result of the evaluation
        """
        ast = exp.getAst() if hasattr(exp, "getAst") else exp

        vars = self.get_expression_variable_values_model(ast, model)

        backup = {}
        for var, value in vars.items():
            backup[var] = self.tt_ctx.getConcreteVariableValue(var)
            self.tt_ctx.setConcreteVariableValue(var, value)
        final_value = ast.evaluate()
        for var in vars.keys():
            self.tt_ctx.setConcreteVariableValue(var, backup[var])
        return final_value

    # def enumerate_expression_value(self, exp: Union[AstNode, Expression], constraints: List[AstNode], values_blacklist: List[int], limit: int):
    #     # Written for when it will work
    #     solver = z3.SolverFor("QF_BV")
    #     ast = exp.getAst() if hasattr(exp, "getAst") else exp
    #     z3ast = self.actx.tritonToZ3(ast)
    #
    #     solver.add([self.actx.tritonToZ3(x) for x in constraints])
    #     solver.add([z3ast != x for x in values_blacklist])
    #
    #     values = []  # retrieved values
    #
    #     while limit:
    #         res = solver.check()
    #         if res == z3.sat:
    #             model = solver.model()
    #             new_val = model.eval(z3ast)
    #             values.append(new_val)
    #             solver.add(z3ast != new_val)
    #         else:
    #             return values
    #         limit -= 1
    #     return values

    def solve_enumerate_expression(self, exp: Union[AstNode, Expression], constraints: List[AstNode], values_blacklist: List[int], limit: int) -> List[Tuple[Model, int]]:
        # Written for when it will work
        ast = exp.getAst() if hasattr(exp, "getAst") else exp

        constraint = self.actx.land(constraints + [ast != x for x in values_blacklist])

        result = []
        while limit:
            status, model = self.solve(constraint, with_pp=False)
            if status == SolverStatus.SAT:
                new_val = self.evaluate_expression_model(ast, model)
                result.append((model, new_val))
                constraint = self.actx.land([constraint, ast != new_val])
            else:
                return result
            limit -= 1
        return result

    @staticmethod
    def from_loader(loader: Loader) -> 'ProcessState':
        pstate = ProcessState(loader.endianness)

        # Initialize the architecture of the processstate
        pstate.initialize_context(loader.architecture)

        # Set the program counter to points to entrypoint
        pstate.cpu.program_counter = loader.entry_point

        # Disable segmentation to map segments
        with pstate.memory.without_segmentation():
            # Load memory areas in memory
            for i, seg in enumerate(loader.memory_segments()):
                if not seg.size and not seg.content:
                    logger.warning(f"A segment have to provide either a size or a content {seg.name} (skipped)")
                    continue
                size = len(seg.content) if seg.content else seg.size
                logger.debug(f"Loading 0x{seg.address:#08x} - {seg.address+size:#08x} size={size:#x}")
                pstate.memory.map(seg.address, size, seg.perms, seg.name)
                if seg.content:
                    pstate.memory.write(seg.address, seg.content)

        # Apply dynamic relocations
        cur_linkage_address = pstate.EXTERN_FUNC_BASE

        # Disable segmentation
        with pstate.memory.without_segmentation():
            # Link imported functions in EXTERN_FUNC_BASE
            for fname, rel_addr in loader.imported_functions_relocations():
                logger.debug(f"Hooking {fname} at {rel_addr:#x}")

                # If we already linked this function (because another library uses it) we reuse the same
                # linkage address.
                if fname in pstate.dynamic_symbol_table:
                    (linkage_address, _) = pstate.dynamic_symbol_table[fname] 
                    logger.debug(f"Already added. {fname} at {rel_addr:#x} linkage_addr={linkage_address:#x}")
                    pstate.memory.write_ptr(rel_addr, linkage_address)

                else:
                    # Add symbol in dynamic_symbol_table
                    pstate.dynamic_symbol_table[fname] = (cur_linkage_address, True)

                    # Apply relocation to our custom address in process memory
                    pstate.memory.write_ptr(rel_addr, cur_linkage_address)
                    # Increment linkage address number
                    cur_linkage_address += pstate.ptr_size

        # Try initializing stack registers if a stack is present in maps
        # Map the stack
        try:
            stack = pstate.memory.map_from_name(pstate.STACK_SEG)
            alloc = 1 * pstate.ptr_size
            pstate.write_register(pstate.base_pointer_register, stack.start+stack.size-alloc) # Pointing right-out of the stack
            pstate.write_register(pstate.stack_pointer_register, stack.start+stack.size-alloc)
        except AssertionError:
            logger.warning("no stack segment has been created by the loader")

        # Search for a map to settle foreign symbols
        segs = pstate.memory.find_map(pstate.EXTERN_SEG)
        if segs:
            symb_base = segs[0].start

            # Link imported symbols
            for sname, rel_addr in loader.imported_variable_symbols_relocations():
                logger.debug(f"Hooking {sname} at {rel_addr:#x}")

                if pstate.architecture == Architecture.X86_64:  # HACK: Keep rel_addr to directly write symbol on it
                    # Add symbol in dynamic_symbol_table
                    pstate.dynamic_symbol_table[sname] = (rel_addr, False)
                    #pstate.memory.write_ptr(rel_addr, cur_linkage_address)  # Do not write anything as symbolic executor will do it
                else:
                    # Add symbol in dynamic_symbol_table
                    pstate.dynamic_symbol_table[sname] = (symb_base, False)
                    pstate.memory.write_ptr(rel_addr, symb_base)

                symb_base += pstate.ptr_size

        for reg_name in pstate.cpu:
            if reg_name in loader.cpustate:
                setattr(pstate.cpu, reg_name, loader.cpustate[reg_name])

        if loader.arch_mode: # If the processor's mode is provided
            if loader.arch_mode == ArchMode.THUMB:
                pstate.set_thumb(True)
        return pstate



from typing import Generator, Optional, Tuple
from pathlib import Path
import logging

# Third-party imports
import cle

# Local imports
from tritondse.loaders import Loader, LoadableSegment
from tritondse.types import Addr, Architecture, PathLike, Platform, Perm, Endian
from tritondse.routines import SUPPORTED_ROUTINES
import tritondse.logging

logger = tritondse.logging.get("loader")

_arch_mapper = {
    "ARMEL":   Architecture.ARM32,
    "AARCH64": Architecture.AARCH64,
    "AMD64":   Architecture.X86_64,
    "X86":   Architecture.X86,
}

_plfm_mapper = {
    "UNIX - System V": Platform.LINUX,
    "windows": Platform.WINDOWS,
    "macos": Platform.MACOS
}


class CleLoader(Loader):
    EXTERN_SYM_BASE = 0x0f001000
    EXTERN_SYM_SIZE = 0x1000

    BASE_STACK = 0xf0000000
    END_STACK = 0x70000000  # This is inclusive

    def __init__(self, path: PathLike, ld_path: Optional[PathLike] = None):
        super(CleLoader, self).__init__(path)
        self.path: Path = Path(path)  #: Binary file path
        if not self.path.is_file():
            raise FileNotFoundError(f"file {path} not found (or not a file)")

        self._disable_vex_loggers()  # disable logs of pyvex

        self.ld_path = ld_path if ld_path is not None else ()
        self.ld = cle.Loader(str(path), ld_path=self.ld_path)

    def _disable_vex_loggers(self):
        for name, logger in logging.root.manager.loggerDict.items():
            if "pyvex" in name:
                logger.propagate = False

    @property
    def name(self) -> str:
        """ Name of the loader"""
        return f"CleLoader({self.path})"

    @property
    def architecture(self) -> Architecture:
        """
        Architecture enum representing program architecture.

        :rtype: Architecture
        """
        return _arch_mapper[self.ld.main_object.arch.name]

    @property
    def endianness(self) -> Endian:
        # FIXME: Depending on architecture returning good endianess
        return Endian.LITTLE

    @property
    def entry_point(self) -> Addr:
        """
        Program entrypoint address as defined in the binary headers

        :rtype: :py:obj:`tritondse.types.Addr`
        """
        return self.ld.main_object.entry

    def memory_segments(self) -> Generator[LoadableSegment, None, None]:
        """
        :return: Generator of tuples addrs and content
        """
        for obj in self.ld.all_objects:
            logger.debug(obj)
            for seg in obj.segments:
                segdata = self.ld.memory.load(seg.vaddr, seg.memsize)
                assert len(segdata) == seg.memsize
                perms = (Perm.R if seg.is_readable else 0) | (Perm.W if seg.is_writable else 0) | (Perm.X if seg.is_executable else 0) 
                if seg.__class__.__name__ != "ExternSegment":
                    # The format string in CLE is broken if the filesize is 0. This is a workaround.
                    logger.debug(f"Loading segment {seg} - perms:{perms}")
                yield LoadableSegment(seg.vaddr, perms, content=segdata, name=f"seg-{obj.binary_basename}")
        # Also return a specific map to put external symbols
        yield LoadableSegment(self.EXTERN_SYM_BASE, self.EXTERN_SYM_SIZE, Perm.R | Perm.W, name="[extern]")
        yield LoadableSegment(self.END_STACK, self.BASE_STACK-self.END_STACK+1, Perm.R | Perm.W, name="[stack]")

        # FIXME. Temporary solution to prevent crashes on access to the TLB e.g fs:28
        yield LoadableSegment(0, 0x2000, Perm.R | Perm.W, name="[fs]")

    # FIXME. Temporary solution to prevent crashes on access to the TLB e.g fs:28
    @property
    def cpustate(self):
        # NOTE: in Triton, the segment selector is used as the segment base and not as a selector into GDT.
        # i.e directly store the segment base into fs
        return {"fs": 0x1000}

    @property
    def platform(self) -> Optional[Platform]:
        """
        Platform of the binary.

        :return: Platform
        """
        return _plfm_mapper[self.ld.main_object.os]

    def imported_functions_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        """
        Iterate over all imported functions by the program. This function
        is a generator of tuples associating the function and its relocation
        address in the binary.

        :return: Generator of tuples function name and relocation address
        """
        # TODO I think there's a problem here. We only deal with imports from the main binary
        # For example if a library calls a libc function, we probably need to patch the library's GOT
        for obj in self.ld.all_objects:
            for fun in obj.imports:
                rtn_name = f"rtn_{fun}"
                if fun in SUPPORTED_ROUTINES:
                    reloc = obj.imports[fun]
                    got_entry_addr = reloc.relative_addr + obj.mapped_base
                    yield fun, got_entry_addr

        # Handle indirect functions.
        # Currently we only support indirect functions if there exists a stub for them in `routines.py`
        # Otherwise the program will crash because CLE doesn't perform the relocation for indirect functions.
        
        # We could perform the relocation ourself by writing to the got slot but we need a way to figure out 
        # the correct fptr to use.
        # In other words we should execute `resolver_fun` or parse it in some way to get the correct function ptr
        # to write to got_slot (write with self.ld.memory.pack_word(got_slot, func_ptr))
        for obj in self.ld.all_objects:
            for (resolver_func, got_rva) in obj.irelatives:
                got_slot = got_rva + obj.mapped_base
                sym = self.ld.find_symbol(resolver_func)
                if sym is None:
                    continue
                fun = sym.name
                if fun in SUPPORTED_ROUTINES:
                    yield fun, got_slot

    def imported_variable_symbols_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        """
        Iterate over all imported variable symbols. Yield for each of them the name and
        the relocation address in the binary.

        :return: Generator of tuples with symbol name, relocation address
        """
        # TODO I think there's a problem here. We only deal with imports from the main binary
        for s in self.ld.main_object.symbols:
            if s.resolved and s._type == cle.SymbolType.TYPE_OBJECT:
                logger.debug(f"CleLoader: hooking symbol {s.name} @ {s.relative_addr:#x} {s.resolved} {s.resolvedby} {s._type}")
                s_addr = s.relative_addr + self.ld.main_object.mapped_base
                yield s.name, s_addr

    def find_function_addr(self, name: str) -> Optional[Addr]:
        """
        Search for the function name in fonctions of the binary.

        :param name: Function name
        :type name: str
        :return: Address of function if found
        :rtype: Addr
        """
        res = [x for x in self.ld.find_all_symbols(name) if x.is_function]
        return res[0].rebased_addr if res else None  # if multiple elements return the first



from __future__ import annotations

from triton import Instruction
from tritondse.callbacks import CbType, ProbeInterface
from tritondse.seed import Seed, SeedStatus
from tritondse.types import Architecture, Addr, Tuple, SolverStatus
from tritondse import SymbolicExecutor, ProcessState
from tritondse.exception import ProbeException
import tritondse.logging

logger = tritondse.logging.get("sanitizers")



def mk_new_crashing_seed(se, model) -> Seed:
    """
    This function is used by every sanitizers to dump the model found in order
    to trigger a bug into the crash directory.

    :return: A fresh Seed
    """
    new_input = bytearray(se.seed.content)
    for k, v in model.items():
        new_input[k] = v.getValue()
    # Don't tag the seed as CRASH before executing it.
    # At this stage, we do not know if the seed will really make the
    # program crash or not.
    return Seed(bytes(new_input))


class UAFSanitizer(ProbeInterface):
    """
    Use-After-Free Sanitizer.
    It is able to detect UaF and double-free. It works by hooking
    all memory read/write if it points to the heap in a freed area
    then the Use-After-Free is detected. It also hooks the free
    routine to detect double-free.
    """
    def __init__(self):
        super(UAFSanitizer, self).__init__()
        self._add_callback(CbType.MEMORY_READ, self._memory_read)
        self._add_callback(CbType.MEMORY_WRITE, self._memory_write)
        self._add_callback(CbType.PRE_RTN, self._free_routine, 'free')

    @staticmethod
    def check(se: SymbolicExecutor, pstate: ProcessState, ptr: Addr, description: str = None) -> bool:
        """
        Checks whether the given ``ptr`` is symptomatic of a Use-After-Free by querying
        various methods of :py:obj:`tritondse.heap_allocator.HeapAllocator`.

        :param se: symbolic executor
        :type se: SymbolicExecutor
        :param pstate: process state
        :type pstate: ProcessState
        :param ptr: pointer address to check
        :type ptr: :py:obj:`tritondse.types.Addr`
        :param description: description string printed in logger if an issue is detected
        :return: True if the bug is present
        """
        if pstate.is_heap_ptr(ptr) and pstate.heap_allocator.is_ptr_freed(ptr):
            if description:
                logger.critical(description)
            se.seed.status = SeedStatus.CRASH
            pstate.stop = True
            return True
        return False

    @staticmethod
    def _memory_read(se, pstate, mem):
        return UAFSanitizer.check(se, pstate, mem.getAddress(), f'UAF detected at {mem}')

    @staticmethod
    def _memory_write(se, pstate, mem, value):
        return UAFSanitizer.check(se, pstate, mem.getAddress(), f'UAF detected at {mem}')

    @staticmethod
    def _free_routine(se, pstate, name, addr):
        ptr = se.pstate.get_argument_value(0)
        return UAFSanitizer.check(se, pstate, ptr, f'Double free detected at {addr:#x}')


class NullDerefSanitizer(ProbeInterface):
    """
    Null Dereference Sanitizer.
    Simply checks if any memory read or write is performed at address 0.
    If so an error is raised.
    """
    def __init__(self):
        super(NullDerefSanitizer, self).__init__()
        self._add_callback(CbType.MEMORY_READ, self._memory_read)
        self._add_callback(CbType.MEMORY_WRITE, self._memory_write)

    @staticmethod
    def check(se: SymbolicExecutor, pstate: ProcessState, ptr: Addr, description: str = None) -> bool:
        """
        Checks that the ``ptr`` given is basically not 0.

        :param se: symbolic executor
        :type se: SymbolicExecutor
        :param pstate: process state
        :type pstate: ProcessState
        :param ptr: pointer address to check
        :type ptr: :py:obj:`tritondse.types.Addr`
        :param description: description string printed in logger if an issue is detected
        :return: True if the bug is present
        """

         # The execution has not started yet
        if pstate.current_instruction is None:
            return False

        # FIXME: Takes so much time...
        #if access_ast is not None and access_ast.isSymbolized():
        #    model = pstate.tt_ctx.getModel(access_ast == 0)
        #    if model:
        #        logging.warning(f'Potential null deref when reading at {mem}')
        #        crash_seed = mk_new_crashing_seed(se, model)
        #        se.workspace.save_seed(crash_seed)
        #        se.seed.status = SeedStatus.OK_DONE
        #        # Do not abort, just continue the execution
        #if access_ast is not None and access_ast.evaluate() == 0:

        # FIXME: Ici on rajoute 16 car nous avons un problème si une instruction se situe
        # en fin de page mappée. Lors du fetching des opcodes, nous fetchons 16 bytes car
        # nous ne connaissons pas la taille d'une instruction, ici, en fetchant en fin de
        # page on déclenche ce sanitizer...

        # FIXME: Why do we call is_valid_memory_mapping ? It is not a "Null Deref vulnerability", it is more a segmentation error
        if ptr == 0 or (pstate.memory.segmentation_enabled and not pstate.memory.is_mapped(ptr)):
            if description:
                logger.critical(description)
            se.seed.status = SeedStatus.CRASH

            # An exception is needed here to break execution on the first
            # invalid memory access. Otherwise, the memory access callback
            # will report on all bytes.
            raise ProbeException(description)

        return False

    @staticmethod
    def _memory_read(se, pstate, mem):
        return NullDerefSanitizer.check(se, pstate, mem.getAddress(), f'Invalid memory access when reading at {mem} from {pstate.current_instruction}')

    @staticmethod
    def _memory_write(se, pstate, mem, value):
        return NullDerefSanitizer.check(se, pstate, mem.getAddress(), f'Invalid memory access when writting at {mem} from {pstate.current_instruction}')


class FormatStringSanitizer(ProbeInterface):
    """
    Format String Sanitizer.
    This probes hooks standard libc functions like 'printf', 'fprintf', 'sprintf',
    'dprintf', 'snprintf' and if one of them is triggered it checks the format string.
    If the format string is symbolic then it is user controlled. A warning is shown
    but the execution not interrupted. However the sanitizer tries through SMT to
    generate format strings with many '%s'. If satisfiable a new input is generated
    which will then be added to inputs to process. That subsequent input might lead
    to a crash.
    """
    def __init__(self):
        super(FormatStringSanitizer, self).__init__()
        self._add_callback(CbType.PRE_RTN,  self._xprintf_arg0_routine, 'printf')
        self._add_callback(CbType.PRE_RTN, self._xprintf_arg1_routine, 'fprintf')
        self._add_callback(CbType.PRE_RTN, self._xprintf_arg1_routine, 'sprintf')
        self._add_callback(CbType.PRE_RTN, self._xprintf_arg1_routine, 'dprintf')
        self._add_callback(CbType.PRE_RTN, self._xprintf_arg1_routine, 'snprintf')

    @staticmethod
    def check(se, pstate, fmt_ptr, extra_data: Tuple[str, Addr] = None):
        """
        Checks that the format string at ``fmt_ptr`` does not contain
        symbolic bytes. If so shows an alert and tries to generate new
        inputs with as many '%s' as possible.

        :param se: symbolic executor
        :type se: SymbolicExecutor
        :param pstate: process state
        :type pstate: ProcessState
        :param fmt_ptr: pointer address to check
        :type fmt_ptr: :py:obj:`tritondse.types.Addr`
        :param extra_data: additionnal infos given by the callbacks on routines (indicating function address)
        :type extra_data: Tuple[str, :py:obj:`tritondse.types.Addr`]
        :return: True if the bug is present
        """
        symbolic_cells = []

        # Count the number of cells which is symbolic
        cur_ptr = fmt_ptr
        while se.pstate.memory.read_uchar(cur_ptr):  # while different from 0
            if se.pstate.is_memory_symbolic(cur_ptr, 1):
                symbolic_cells.append(cur_ptr)
            cur_ptr += 1

        if symbolic_cells:
            extra = f"(function {extra_data[0]}@{extra_data[1]:#x})" if extra_data else ""
            logger.warning(f'Potential format string of length {len(symbolic_cells)} on {fmt_ptr:x} {extra}')
            se.seed.status = SeedStatus.OK_DONE
            pp_seeds = []
            nopp_seeds = []

            for i in range(int(len(symbolic_cells) / 2)):
                # FIXME: Does not check that cell1 and cell2 are contiguous
                cell1 = pstate.read_symbolic_memory_byte(symbolic_cells.pop(0)).getAst()
                cell2 = pstate.read_symbolic_memory_byte(symbolic_cells.pop(0)).getAst()

                # Try to solve once with the path predicate
                st, model = pstate.solve([cell1 == ord('%'), cell2 == ord('s')], with_pp=True)
                if st == SolverStatus.SAT and model:
                    pp_seeds.append(mk_new_crashing_seed(se, model))

                # Try once again without the path predicate (may be incorrect but help to discover bug)
                st, model = pstate.solve_no_pp([cell1 == ord('%'), cell2 == ord('s')])
                if st == SolverStatus.SAT and model:
                    pp_seeds.append(mk_new_crashing_seed(se, model))

            # If found some seeds
            if pp_seeds:
                s = pp_seeds[-1]
                se.enqueue_seed(s)  # Only keep last seed
                logger.warning(f'Found model that might lead to a crash: {s.hash} (with path predicate)')
            if nopp_seeds:
                s = nopp_seeds[-1]
                se.enqueue_seed(s)  # Only keep last seed
                logger.warning(f'Found model that might lead to a crash: {s.hash} (without path predicate)')

            # Do not stop the execution, just continue the execution
            pstate.stop = False
            return True
        return False

    @staticmethod
    def _xprintf_arg0_routine(se, pstate, name, addr):
        string_ptr = se.pstate.get_argument_value(0)
        FormatStringSanitizer.check(se, pstate, string_ptr, (name, addr))

    @staticmethod
    def _xprintf_arg1_routine(se, pstate, name, addr):
        string_ptr = se.pstate.get_argument_value(1)
        FormatStringSanitizer.check(se, pstate, string_ptr, (name, addr))


class IntegerOverflowSanitizer(ProbeInterface):
    """
    Integer Overflow Sanitizer.
    This probe checks on every instructions that the overflow
    flag is not set. If so mark the input as a crashing input.
    If not, but the value is symbolic, via SMT solving to make
    it to be set (and thus to overflow). If possible generates
    a new input to be executed.
    """
    def __init__(self):
        super(IntegerOverflowSanitizer, self).__init__()
        self._add_callback(CbType.POST_INST, self.check)

    @staticmethod
    def check(se: SymbolicExecutor, pstate: ProcessState, instruction: Instruction) -> bool:
        """
        The entry point of the sanitizer. This function check if a bug is present

        :param se: symbolic executor
        :type se: SymbolicExecutor
        :param pstate: process state
        :type pstate: ProcessState
        :param instruction: Instruction that has just been executed
        :type instruction: `Instruction <https://triton.quarkslab.com/documentation/doxygen/py_Instruction_page.html>`_
        :return: True if the bug is present
        """
        # This probe is only available for X86_64 and AARCH64
        assert(pstate.architecture == Architecture.X86_64 or pstate.architecture == Architecture.AARCH64)

        rf = (pstate.registers.of if pstate.architecture == Architecture.X86_64 else pstate.registers.v)

        if pstate.read_register(rf):
            logger.warning(f'Integer overflow at {instruction}')
            # FIXME: What if it's normal behavior?
            se.seed.status = SeedStatus.CRASH
            return True

        else:  # if no overflow took place check if symbolic and if so, if it can be 1
            if pstate.is_register_symbolic(rf):
                sym_flag = pstate.read_symbolic_register(rf)
                _, model = pstate.solve_no_pp(sym_flag.getAst() == 1)
                if model:
                    logger.warning(f'Potential integer overflow at {instruction}')
                    crash_seed = mk_new_crashing_seed(se, model)
                    se.enqueue_seed(crash_seed)
                    return True

        return False



import logging

'''
Loggers hierarchy is the following:

- tritondse.
    - 

'''


logger = logging.getLogger('tritondse')
logger.propagate = False  # Do not propagate logs by default
color_enabled = True
_loggers = {}

def get(name: str = "") -> logging.Logger:
    """
    Get a child logger from the tritondse one.
    If name is empty returns the root tritondse
    logger.

    :param name: logger name
    """
    log = logger.getChild(name) if name else logger

    if log.name not in _loggers:
        log.propagate = False  # first time it is retrieve disable propagation
        _loggers[log.name] = log

    return log


def enable(level: int = logging.INFO, name: str = "") -> None:
    """
    Enable tritondse logging to terminal output

    :param level: logging level
    :param name: name of the logger to enable (all by default)
    """
    log = get(name)
    log.propagate = True
    log.setLevel(level)

    # Enable root logger if needed
    if log.name != "tritondse":
        logger.propagate = True
    else:
        for sub_logger in _loggers.values():
            sub_logger.propagate = True


def enable_to_file(level: int, file: str, name: str = "") -> None:
    """
    Enable tritondse logging to a file

    :param level: logging level
    :param file: path to log file
    :param name: name of the logger to enable to a file
    """
    log = get(name)
    log.setLevel(level)
    fmt = logging.Formatter("%(asctime)s %(threadName)s [%(levelname)s] %(message)s")

    handler = logging.FileHandler(file)
    handler.setFormatter(fmt)
    log.addHandler(handler)  # Add the handler to the logger


# built-in imports
from __future__ import annotations
import hashlib
import struct
from pathlib import Path
from typing import List, Generator, Tuple, Set, Union, Dict, Optional
from collections import Counter
from enum import IntFlag, Enum, auto
import pickle
import enum_tools.documentation

# third-party imports
from triton import AST_NODE

# local imports
from tritondse.types import Addr, PathConstraint, PathBranch, SolverStatus, PathHash, Edge, SymExType
import tritondse.logging

logger = tritondse.logging.get()

CovItem = Union[Addr, Edge, PathHash, Tuple[PathHash, Edge]]
"""
Variant type representing a coverage item.
It can be:

* an address :py:obj:`tritondse.types.Addr` for block coverage
* an edge :py:obj:`tritondse.types.Edge` for edge coverage
* a string :py:obj:`tritondse.types.PathHash` for path coverage
* a tuple of both a Pathhash and an edge
"""

@enum_tools.documentation.document_enum
class CoverageStrategy(str, Enum):
    """
    Coverage strategy (metric) enum.
    This enum will change whether a given branch have
    to be solved or not.
    """

    BLOCK = "block"   # doc: block coverage, only tracks new basic blocks covered
    EDGE = "edge"     # doc: edge coverage, tracks CFGs edges covered
    PATH = "path"     # doc: tracks any new path covered
    PREFIXED_EDGE = "PREFIXED_EDGE"  # doc: edge coverage but also taking in account path prefix)


@enum_tools.documentation.document_enum
class BranchSolvingStrategy(IntFlag):
    """
    Branch strategy enumerate.
    It defines the manner with which branches are checked with SMT
    on a single trace, namely a :py:obj:`CoverageSingleRun`. For a
    given branch that has not been covered strategies are:

    * ``ALL_NOT_COVERED``: check by SMT all occurences
    * ``FIRST_LAST_NOT_COVERED``: check only the first and last occurence in the trace
    """
    ALL_NOT_COVERED = auto()         # doc: check by SMT all occurences of a given branch (true by default)
    FIRST_LAST_NOT_COVERED = auto()  # doc: check by SMT the first and last occurence of a given branch
    UNSAT_ONCE = auto()              # doc: if a branch is UNSAT do not try solving it again
    TIMEOUT_ONCE = auto()            # doc: if a branch is TIMEOUT do not try solving it again
    TIMEOUT_ALWAYS = auto()          # doc: always try solving again a TIMEOUT branch (incompatible with :py:enum:mem:`TIMEOUT_ONCE`
    COVER_SYM_DYNJUMP = auto()       # doc: try covering dynamic jumps on a symbolic register or memory value
    COVER_SYM_READ = auto()          # doc: try enumerating values for symbolic reads
    COVER_SYM_WRITE = auto()         # doc: try enumerating values for symbolic writes
    SOUND_MEM_ACCESS = auto()        # doc: enables adding a constraint when using a symbolic read/write or jump
    MANUAL = auto()                  # doc: disable automatic branch solving after an execution (has to be done manually in callbacks)


class CoverageSingleRun(object):
    """
    Coverage produced by a **Single Execution**
    Depending on the strategy given to the constructor
    it stores different data.
    """

    def __init__(self, strategy: CoverageStrategy):
        """
        :param strategy: Strategy to employ
        :type strategy: CoverageStrategy
        """
        self.strategy: CoverageStrategy = strategy  #: Coverage strategy

        # For instruction coverage
        self.covered_instructions: Dict[Addr, int] = Counter()
        """ Instruction coverage. Counter for code coverage) """

        self.covered_items: Dict[CovItem, int] = Counter()
        """ Stores covered items whatever they are """
        self.not_covered_items: Set[CovItem] = set()
        self._not_covered_items_mirror: Dict[CovItem, List[str]] = {}  # solely used for prefixed-edge
        """ CovItems not covered in the trace. It thus represent what can be
        covered by the trace (input). We call it coverage objectives."""

        # For path coverage
        self._current_path: List[Addr] = []
        """ List of addresses forming the path currently being taken """
        self._current_path_hash = hashlib.md5()

    def add_covered_address(self, address: Addr) -> None:
        """
        Add an instruction address covered.
        *(Called by :py:obj:`SymbolicExecutor` for each
        instruction executed)*

        :param address: The address of the instruction
        :type address: :py:obj:`tritondse.types.Addr`
        """
        self.covered_instructions[address] += 1

    def add_covered_dynamic_branch(self, source: Addr, target: Addr) -> None:
        """
        Add a dynamic branch covered. The branch will be encoded according to the
        coverage strategy.

        :param source: Address of the dynamic jump
        :param target: Target address on which the jump is performed
        :return:
        """
        if self.strategy == CoverageStrategy.BLOCK:
            pass # Target address will be covered anyway

        if self.strategy == CoverageStrategy.EDGE:
            self.covered_items[(source, target)] += 1
            self.not_covered_items.discard((source, target))    # Remove it from non-taken if it was inside

        if self.strategy == CoverageStrategy.PATH:
            self._current_path.append(target)
            self._current_path_hash.update(struct.pack("<Q", target))
            self.covered_items[self._current_path_hash.hexdigest()] += 1

        if self.strategy == CoverageStrategy.PREFIXED_EDGE:
            # Add covered as covered
            self.covered_items[("", (source, target))] += 1
            # update the current path hash etc
            self._current_path.append(target)
            self._current_path_hash.update(struct.pack("<Q", target))

    def add_covered_branch(self, program_counter: Addr, taken_addr: Addr, not_taken_addr: Addr) -> None:
        """
        Add a branch to our covered branches list. Each branch is encoded according
        to the coverage strategy. For code coverage, the branch encoding is the
        address of the instruction. For edge coverage, the branch encoding is the
        tupe (src address, dst address). For path coverage, the branch encoding
        is the MD5 of the conjunction of all taken branch addresses.

        :param program_counter: The address taken in by the branch
        :type program_counter: :py:obj:`tritondse.types.Addr`
        :param taken_addr: Target address of branch taken
        :type taken_addr: Addr
        :param not_taken_addr: Target address of branch **not** taken
        :type not_taken_addr: Addr
        """

        if self.strategy == CoverageStrategy.BLOCK:
            self.covered_items[taken_addr] += 1
            self.not_covered_items.discard(taken_addr)    # remove address from non-covered if inside
            if not_taken_addr not in self.covered_items:  # Keep the address that has not been covered (and could have)
                self.not_covered_items.add(not_taken_addr)

        if self.strategy == CoverageStrategy.EDGE:
            taken_tuple, not_taken_tuple = (program_counter, taken_addr), (program_counter, not_taken_addr)
            self.covered_items[taken_tuple] += 1
            self.not_covered_items.discard(taken_tuple)    # Remove it from non-taken if it was inside
            if not_taken_tuple not in self.covered_items:  # Add the not taken tuple in non-covered
                self.not_covered_items.add(not_taken_tuple)

        if self.strategy == CoverageStrategy.PATH:
            self._current_path.append(taken_addr)

            # Compute the hash of the not taken path and add it to non-covered paths
            not_taken_path_hash = self._current_path_hash.copy()
            not_taken_path_hash.update(struct.pack('<Q', not_taken_addr))
            self.not_covered_items.add(not_taken_path_hash.hexdigest())

            # Update the current path hash and add it to hashes
            self._current_path_hash.update(struct.pack("<Q", taken_addr))
            self.covered_items[self._current_path_hash.hexdigest()] += 1

        if self.strategy == CoverageStrategy.PREFIXED_EDGE:
            taken_tuple, not_taken_tuple = (program_counter, taken_addr), (program_counter, not_taken_addr)
            _, not_taken = (self._current_path_hash.hexdigest(), taken_tuple), (self._current_path_hash.hexdigest(), not_taken_tuple)
            gtaken, gnot_taken = ("", taken_tuple), ("", not_taken_tuple)

            # Add covered as covered
            self.covered_items[gtaken] += 1

            # Find all items in not_covered that have this edge
            if taken_tuple in self._not_covered_items_mirror:               # if a not_covered match this edge
                for prefix in self._not_covered_items_mirror[taken_tuple]:  # iterate all the prefixes
                    self.not_covered_items.discard((prefix, taken_tuple))   # and discard them
                self._not_covered_items_mirror.pop(taken_tuple)             # finally discard the entry

            # look if not_taken edge not in covered
            if gnot_taken not in self.covered_items:
                self.not_covered_items.add(not_taken)
                if not_taken[1] not in self._not_covered_items_mirror:
                    self._not_covered_items_mirror[not_taken[1]] = [not_taken[0]]
                else:
                    self._not_covered_items_mirror[not_taken[1]].append(not_taken[0])

            # update the current path hash etc
            self._current_path.append(taken_addr)
            self._current_path_hash.update(struct.pack("<Q", taken_addr))

    @property
    def unique_instruction_covered(self) -> int:
        """
        :return: The number of unique instructions covered
        """
        return len(self.covered_instructions)

    @property
    def unique_covitem_covered(self) -> int:
        """
        :return: The number of unique edges covered
        """
        return len(self.covered_items)

    @property
    def total_instruction_executed(self) -> int:
        """
        :return: The number of total instruction executed
        """
        return sum(self.covered_instructions.values())

    def post_execution(self) -> None:
        """
        Function is called after each execution
        for post processing or clean-up. *(Not
        doing anythin at the moment)*
        """
        pass

    def is_covered(self, item: CovItem) -> bool:
        """
        Return whether the item has been covered or not.
        **The item should match the strategy**

        :param item: An address, an edge or a path
        :type item: CovItem
        :return: bool
        """
        if self.strategy == CoverageStrategy.PREFIXED_EDGE:
            try:
                return ('', item[1]) in self.covered_items
            except TypeError:  # in case of ellipsis
                return False
        else:
            return item in self.covered_items

    def pp_item(self, covitem: CovItem) -> str:
        """
        Pretty print a CovItem according the coverage strategy

        :param covitem: An address, an edge or a path
        :return: str
        """
        if self.strategy == CoverageStrategy.BLOCK:
            return f"0x{covitem:08x}"
        elif self.strategy == CoverageStrategy.EDGE:
            return f"(0x{covitem[0]:08x}-0x{covitem[1]:08x})"
        elif self.strategy == CoverageStrategy.PATH:
            return covitem[:10]  # already a hash str
        elif self.strategy == CoverageStrategy.PREFIXED_EDGE:
            return f"({covitem[0][:6]}_0x{covitem[1][0]:08x}-0x{covitem[1][1]:08x})"

    def difference(self, other: CoverageSingleRun) -> Set[CovItem]:
        if self.strategy == other.strategy:
            return self.covered_items.keys() - other.covered_items.keys()
        else:
            logger.error("Trying to make difference of coverage with different strategies")
            return set()

    def __sub__(self, other) -> Set[CovItem]:
        return self.difference(other)


class GlobalCoverage(CoverageSingleRun):
    """
    Global Coverage.
    Represent the overall coverage of the exploration.
    It is filled by iteratively call merge with the
    :py:obj:`CoverageSingleRun` objects created during
    exploration.
    """

    COVERAGE_FILE = "coverage.json"

    def __init__(self, strategy: CoverageStrategy, branch_strategy: BranchSolvingStrategy):
        """
        :param strategy: Coverage strategy to use
        :type strategy: CoverageStrategy
        :param branch_strategy: Branch checking strategies
        :type branch_strategy: BranchSolvingStrategy
        """
        super().__init__(strategy)
        self.branch_strategy = branch_strategy

        # Keep pending items to be covered (code, edge, path)
        self.pending_coverage: Set[CovItem] = set()
        """ Set of pending coverage items. These are items for which a branch
        as already been solved and 
        """

        self.uncoverable_items: Dict[CovItem, SolverStatus] = {}
        """ CovItems that are determined not to be coverable. """

        self.covered_symbolic_pointers: Set[Addr] = set()
        """ Set of addresses for which pointers have been enumerated """

    def iter_new_paths(self, path_constraints: List[PathConstraint]) -> Generator[Tuple[SymExType, List[PathConstraint], PathBranch, CovItem, int], Optional[SolverStatus], None]:
        """
        The function iterate the given path predicate and yield PatchConstraint to
        consider as-is and PathBranch representing the new branch to take. It acts
        as a black-box so that the SeedManager does not have to know what strategy
        is being used under the hood. From an implementation perspective the goal
        of the function is to manipulate the path WITHOUT doing any SMT related things.

        :param path_constraints: list of path constraint to iterate
        :return: generator of path constraint and branches to solve. The first tuple
                 item is a list of PathConstraint to add in the path predicate and the second
                 is the branch to solve (but not to keep in path predicate)
        """
        if BranchSolvingStrategy.MANUAL in self.branch_strategy:
            logger.info(f'Branch solving strategy set to MANUAL.')
            return

        pending_csts = []
        current_hash = hashlib.md5()  # Current path hash for PATH coverage

        # NOTE: When we arrive here the CoverageSingleRun associated with the path_constraints
        # has already been merge. Thus covered, pending etc do include ones of the CoverageSingleRuns

        not_covered_items = self._get_items_trace(path_constraints)  # Map of CovItem -> [idx1, idx2, ..., .] (occurence in list)
        # is_ok_with_branch_strategy = lambda covitem, idx: True if self.strategy == CoverageStrategy.PATH else (idx in occurence_map[covitem])

        # Re-iterate through all path constraints to solve them concretely (with knowledge of what is beyond in the trace)
        for i, pc in enumerate(path_constraints):
            if pc.isMultipleBranches():     # If there is a condition
                for branch in pc.getBranchConstraints():  # Get all branches
                    # Get the constraint of the branch which has not been taken.
                    if not branch['isTaken']:
                        covitem = self._get_covitem(current_hash, branch)
                        generic_covitem = ('', covitem[1]) if self.strategy == CoverageStrategy.PREFIXED_EDGE else covitem
                        #print(f"Covitem: {covitem}: {covitem not in self.covered_items} | {covitem not in self.pending_coverage} | {covitem not in self.uncoverable_items} | {i in not_covered_items.get(covitem, [])} | {i} | {not_covered_items.get(covitem)}")

                        # Not covered in: previous runs | yet to be covered by a seed already SAT | not uncoverable | parts of items to solve
                        if generic_covitem not in self.covered_items and \
                           generic_covitem not in self.pending_coverage and \
                           covitem not in self.uncoverable_items and \
                           i in not_covered_items.get(covitem, []):

                            # Send the branch to solve to the function iterating
                            res = yield SymExType.CONDITIONAL_JMP, pending_csts, branch, covitem, i

                            # If path SAT add it to pending coverage
                            if res == SolverStatus.SAT:
                                self.pending_coverage.add(generic_covitem)

                            elif res == SolverStatus.UNSAT:
                                if BranchSolvingStrategy.UNSAT_ONCE in self.branch_strategy:
                                    self.uncoverable_items[covitem] = res
                                elif self.strategy in [CoverageStrategy.PATH, CoverageStrategy.PREFIXED_EDGE]:
                                    self.uncoverable_items[covitem] = res  # paths, and prefixed-edge ensure to be unique thus drop if unsat

                            elif res == SolverStatus.TIMEOUT:
                                if BranchSolvingStrategy.TIMEOUT_ONCE in self.branch_strategy:
                                    self.uncoverable_items[covitem] = res

                            elif res == SolverStatus.UNKNOWN:
                                pass

                            else: # status == None
                                logger.debug(f'Branch skipped!')

                            pending_csts = []  # reset pending constraint added

                    else:
                        pass  # Branch was taken do nothing

                # Add it the path predicate constraints and update current path hash
                pending_csts.append(pc)
                current_hash.update(struct.pack("<Q", pc.getTakenAddress()))

            else:
                cmt = pc.getComment()

                if (cmt.startswith("dyn-jmp") and BranchSolvingStrategy.COVER_SYM_DYNJUMP in self.branch_strategy) or \
                   (cmt.startswith("sym-read") and BranchSolvingStrategy.COVER_SYM_READ in self.branch_strategy) or \
                   (cmt.startswith("sym-write") and BranchSolvingStrategy.COVER_SYM_WRITE in self.branch_strategy):
                    typ, offset, addr = cmt.split(":")
                    typ = SymExType(typ)
                    offset, addr = int(offset), int(addr)
                    if addr not in self.covered_symbolic_pointers:  # if the address pointer has never been covered
                        pred = pc.getTakenPredicate()
                        if pred.getType() == AST_NODE.EQUAL:
                            p1, p2 = pred.getChildren()
                            if p2.getType() == AST_NODE.BV:
                                logger.info(f"Try to enumerate value {offset}:0x{addr:02x}: {p1}")
                                res = yield typ, pending_csts, p1, (addr, p2.evaluate()), i
                                self.covered_symbolic_pointers.add(addr)  # add the pointer in covered regardless of result
                            else:
                                logger.warning(f"memory constraint unexpected pattern: {pred}")
                        else:
                            logger.warning(f"memory constraint unexpected pattern: {pred}")

                    if BranchSolvingStrategy.SOUND_MEM_ACCESS in self.branch_strategy:
                        pending_csts.append(pc)  # if sound add the mem dereference as a constraint in path predicate
                        # NOTE: in both case the branch is not taken in account in the current_path_hash

                else:  # Routines, or user-defined constraints thus add it all the time.
                    pending_csts.append(pc)

    def _get_covitem(self, path_hash, branch: PathBranch) -> CovItem:
        src, dst = branch['srcAddr'], branch['dstAddr']

        # Check if the target is new with regards to the strategy
        if self.strategy == CoverageStrategy.BLOCK:
            return dst
        elif self.strategy == CoverageStrategy.EDGE:
            return src, dst
        elif self.strategy == CoverageStrategy.PATH:
            # Have to fork the hash of the current pc for each branch we want to revert
            forked_hash = path_hash.copy()
            forked_hash.update(struct.pack("<Q", dst))
            return forked_hash.hexdigest()
        elif self.strategy == CoverageStrategy.PREFIXED_EDGE:
            return path_hash.hexdigest(), (src, dst)
        else:
            assert False

    def _get_items_trace(self, path_constraints: List[PathConstraint]) -> Dict[CovItem, List[int]]:
        """
        Iterate the all trace and retrieve all covered and not covered CovItem. For non covered one
        it filter instances to check.
        """
        not_covered = {}
        current_hash = hashlib.md5()  # Current path hash for PATH coverage
        for i, pc in enumerate(path_constraints):
            if pc.isMultipleBranches():     # If there is a condition
                for branch in pc.getBranchConstraints():  # Get all branches
                    if not branch['isTaken']:
                        covitem = self._get_covitem(current_hash, branch)
                        if covitem in not_covered:
                            not_covered[covitem].append(i)
                        else:
                            not_covered[covitem] = [i]
                current_hash.update(struct.pack("<Q", pc.getTakenAddress()))  # compute current path hash along the way
            else:
                pass  # Ignore all other dynamic constraints in path computation

        # Now filter the map according to the branch solving strategy
        if BranchSolvingStrategy.FIRST_LAST_NOT_COVERED in self.branch_strategy:
            if self.strategy == CoverageStrategy.PREFIXED_EDGE:
                # Black magic
                m = {("", e): [] for h, e in not_covered.keys()}  # Map: ("", edge) -> []
                for (h, e), v in not_covered.items():          # fill map with all occurences edges regardless of path
                    m[("", e)].extend(v)
                for k in m.keys():                             # iterate the result and only keep min and max occurence
                    idxs = m[k]
                    if len(idxs) > 2:
                        m[k] = [min(idxs), max(idxs)]
                for k in not_covered.keys():                   # Push back resulting list in not_covered items
                    not_covered[k] = m[('', k[1])]

            else:  # Straightforward
                for k in not_covered.keys():
                    l = not_covered[k]
                    if len(l) > 2:
                        not_covered[k] = [l[0], l[-1]]  # Only keep first and last iteration
        else: # ALL_NOT_COVERED
            pass  # Keep all occurences
        return not_covered


    def merge(self, other: CoverageSingleRun) -> None:
        """
        Merge a CoverageSingeRun instance into this instance

        :param other: The CoverageSingleRun to merge into self
        :type other: CoverageSingleRun
        """
        assert self.strategy == other.strategy

        # Update instruction coverage for code coverage (in all cases keep code coverage)
        self.covered_instructions.update(other.covered_instructions)

        # Remove covered items from pending ones
        self.pending_coverage.difference_update(other.covered_items)

        # Update covered items
        self.covered_items.update(other.covered_items)

        # Update non-covered ones
        if self.strategy == CoverageStrategy.PREFIXED_EDGE:
            # More complex as not_covered as covitem: (hash, edge) while covered has covitems: ("", edge)

            # remove self not covered that are now covered
            for _, edge in other.covered_items.keys():
                if edge in self._not_covered_items_mirror:
                    for prefix in self._not_covered_items_mirror[edge]:  # iterate all the prefixes
                        self.not_covered_items.discard((prefix, edge))  # and discard them
                    self._not_covered_items_mirror.pop(edge)  # finally discard the entry

            # Only add other not covered if still not covered
            for prefix, edge in other.not_covered_items:
                if ("", edge) not in self.covered_items:
                    self.not_covered_items.add((prefix, edge))
                    if edge not in self._not_covered_items_mirror:
                        self._not_covered_items_mirror[edge] = [prefix]
                    else:
                        self._not_covered_items_mirror[edge].append(prefix)

        else: # Straightfoward set difference
            self.not_covered_items.update(other.not_covered_items - self.covered_items.keys())


    def can_improve_coverage(self, other: CoverageSingleRun) -> bool:
        """
        Check if some of the non-covered are not already in the global coverage
        Used to know if an input is relevant to keep or not

        :param other: The CoverageSingleRun to check against our global coverage state
        :return: bool
        """
        return bool(self.new_items_to_cover(other))


    def can_cover_symbolic_pointers(self, execution: 'SymbolicExecutor') -> bool:
        """
        Determines if this execution has symbolic memory accesses to enumerate. If so we may want
        to enumerate them even though 
        """
        path_constraints = execution.pstate.get_path_constraints()

        for pc in path_constraints:
            if not pc.isMultipleBranches():     # If there isn't a condition i.e it's a sym ptr access
                cmt = pc.getComment()

                if (cmt.startswith("dyn-jmp") and BranchSolvingStrategy.COVER_SYM_DYNJUMP in self.branch_strategy) or \
                   (cmt.startswith("sym-read") and BranchSolvingStrategy.COVER_SYM_READ in self.branch_strategy) or \
                   (cmt.startswith("sym-write") and BranchSolvingStrategy.COVER_SYM_WRITE in self.branch_strategy):
                    typ, offset, addr = cmt.split(":")
                    typ = SymExType(typ)
                    offset, addr = int(offset), int(addr)
                    if addr not in self.covered_symbolic_pointers:  # if the address pointer has never been covered
                        return True
        return False


    def new_items_to_cover(self, other: CoverageSingleRun) -> Set[CovItem]:
        """
        Return all coverage items (addreses, edges, paths) that the given CoverageSingleRun
        can cover if it is possible to negate their branches

        :param other: The CoverageSingleRun to check with our global coverage state
        :return: A set of CovItem
        """
        assert self.strategy == other.strategy
        # Take not covered_items (potential candidates) substract already covered items, uncoverable and pending ones.
        # Resulting covitem are really new ones that the trace brings
        return other.not_covered_items - self.covered_items.keys() - self.uncoverable_items.keys() - self.pending_coverage

    def improve_coverage(self, other: CoverageSingleRun) -> bool:
        """
        Checks if the given object do cover new covitem than the current
        coverage. More concretely it performs the difference between the
        two covered dicts. If ``other`` contains new items return True.

        :param other: coverage on which to check coverage
        :return: Whether the coverage covers new items
        """
        return bool(other.covered_items.keys() - self.covered_items.keys())

    @staticmethod
    def from_file(file: Union[str, Path]) -> 'GlobalCoverage':
        with open(file, "rb") as f:
            obj = pickle.load(f)
        return obj

    def to_file(self, file: Union[str, Path]) -> None:
        copy = self._current_path_hash
        self._current_path_hash = None
        with open(file, "wb") as f:
            pickle.dump(self, f)
        self._current_path_hash = copy


    def post_exploration(self, workspace: 'Workspace') -> None:
        """ Function called at the very end of the exploration.
        It saves the coverage in the workspace.

        :param workspace: Workspace in which to save coverage
        :type workspace: Workspace
        """
        # Save the coverage
        self.to_file(workspace.get_metadata_file_path(self.COVERAGE_FILE))

    def clone(self) -> 'GlobalCoverage':
        cov2 = GlobalCoverage(self.strategy, self.branch_strategy)

        # Copy items from the CoverageSingleRun
        cov2.covered_instructions = Counter({k: v for k, v in self.covered_instructions.items()})
        cov2.covered_items = Counter({k: v for k, v in self.covered_items.items()})
        cov2.not_covered_items = {x for x in self.not_covered_items}
        cov2._not_covered_items_mirror = {k: v for k, v in self._not_covered_items_mirror.items()}
        cov2._current_path = self._current_path[:]
        self._current_path: List[Addr] = []
        self._current_path_hash = self._current_path_hash.copy()

        # Copy items from the global coverage
        cov2.pending_coverage = {x for x in self.pending_coverage}
        cov2.uncoverable_items = {k: v for k, v in self.uncoverable_items.items()}
        cov2.covered_symbolic_pointers = {x for x in self.covered_symbolic_pointers}
        return cov2
